"""Light HCL rules."""

import abc
import datetime
import logging
import typing

import HABApp

import habapp_rules.actors.config.light_hcl
import habapp_rules.actors.state_observer
import habapp_rules.core.exceptions
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.core.type_of_day
import habapp_rules.system

LOGGER = logging.getLogger(__name__)


class _HclBase(habapp_rules.core.state_machine_rule.StateMachineRule):
    """Base class for HCL rules."""

    states: typing.ClassVar = [
        {"name": "Manual"},
        {"name": "Hand", "timeout": 99, "on_timeout": "hand_timeout"},
        {
            "name": "Auto",
            "initial": "Init",
            "children": [{"name": "Init"}, {"name": "HCL"}, {"name": "Sleep", "initial": "Active", "children": [{"name": "Active"}, {"name": "Post", "timeout": 1, "on_timeout": "post_sleep_timeout"}]}, {"name": "Focus"}],
        },
    ]

    trans: typing.ClassVar = [
        {"trigger": "manual_on", "source": ["Auto", "Hand"], "dest": "Manual"},
        {"trigger": "manual_off", "source": "Manual", "dest": "Auto"},
        {"trigger": "hand_on", "source": "Auto", "dest": "Hand"},
        {"trigger": "hand_timeout", "source": "Hand", "dest": "Auto"},
        {"trigger": "sleep_start", "source": ["Auto_HCL", "Auto_Focus"], "dest": "Auto_Sleep"},
        {"trigger": "sleep_end", "source": "Auto_Sleep_Active", "dest": "Auto_Sleep_Post"},
        {"trigger": "post_sleep_timeout", "source": "Auto_Sleep_Post", "dest": "Auto_HCL"},
        {"trigger": "focus_start", "source": ["Auto_HCL", "Auto_Sleep"], "dest": "Auto_Focus"},
        {"trigger": "focus_end", "source": "Auto_Focus", "dest": "Auto_HCL"},
    ]

    def __init__(self, config: habapp_rules.actors.config.light_hcl.HclElevationConfig | habapp_rules.actors.config.light_hcl.HclTimeConfig) -> None:
        """Init base class.

        Args:
            config: config of HCL light.
        """
        self._config = config

        habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.color.name)

        self._state_observer = habapp_rules.actors.state_observer.StateObserverNumber(self._config.items.color.name, self._cb_hand, value_tolerance=10)

        # init state machine
        self._previous_state = None
        self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(model=self, states=self.states, transitions=self.trans, ignore_invalid_triggers=True, after_state_change="_update_openhab_state")
        self._set_initial_state()

        self._set_timeouts()

        # set callbacks
        self._config.items.manual.listen_event(self._cb_manual, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.sleep_state is not None:
            self._config.items.sleep_state.listen_event(self._cb_sleep, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.focus is not None:
            self._config.items.focus.listen_event(self._cb_focus, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.switch_on is not None:
            self._config.items.switch_on.listen_event(self._cb_switch_on, HABApp.openhab.events.ItemStateChangedEventFilter())

    def _set_timeouts(self) -> None:
        """Set timeouts."""
        self.state_machine.get_state("Auto_Sleep_Post").timeout = self._config.parameter.post_sleep_timeout
        self.state_machine.get_state("Hand").timeout = self._config.parameter.hand_timeout

    def _get_initial_state(self, default_value: str = "") -> str:  # noqa: ARG002
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            if OpenHAB item has a state it will return it, otherwise return the given default value
        """
        if self._config.items.manual.is_on():
            return "Manual"
        if self._config.items.sleep_state is not None and self._config.items.sleep_state.value in {habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value}:
            return "Auto_Sleep"
        if self._config.items.focus is not None and self._config.items.focus.is_on():
            return "Auto_Focus"
        return "Auto_HCL"

    def _update_openhab_state(self) -> None:
        """Update OpenHAB state item and other states.

        This should method should be set to "after_state_change" of the state machine.
        """
        if self.state != self._previous_state:
            super()._update_openhab_state()
            self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

            self._set_light_color()
            self._previous_state = self.state

    def _set_light_color(self) -> None:
        """Set light color."""
        target_color = None

        if self.state == "Auto_HCL":
            target_color = self._get_hcl_color()
        elif self.state == "Auto_Focus":
            target_color = self._config.parameter.focus_color
        elif self.state == "Auto_Sleep_Active":
            target_color = self._config.parameter.sleep_color

        if target_color is not None:
            self._state_observer.send_command(target_color)

    def on_enter_Auto_Init(self) -> None:  # noqa: N802
        """Is called on entering of init state."""
        self._set_initial_state()

    @abc.abstractmethod
    def _get_hcl_color(self) -> int | None:
        """Get HCL color.

        Returns:
            HCL light color
        """

    @staticmethod
    def _get_interpolated_value(config_start: tuple[float, float], config_end: tuple[float, float], value: float) -> float:
        """Get interpolated value.

        Args:
            config_start: start config
            config_end: end config
            value: input value which is the input for the interpolation

        Returns:
            interpolated value
        """
        fit_m = (config_end[1] - config_start[1]) / (config_end[0] - config_start[0])
        fit_t = config_end[1] - fit_m * config_end[0]

        return fit_m * value + fit_t

    def _cb_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the manual switch has a state change event.

        Args:
            event: trigger event
        """
        if event.value == "ON":
            self.manual_on()
        else:
            self.manual_off()

    def _cb_hand(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:  # noqa: ARG002
        """Callback, which is triggered by the state observer if a manual change was detected.

        Args:
            event: original trigger event
        """
        self._instance_logger.debug("Hand detected")
        self.hand_on()

    def _cb_focus(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the focus switch has a state change event.

        Args:
            event: trigger event
        """
        if event.value == "ON":
            self.focus_start()
        else:
            self.focus_end()

    def _cb_sleep(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the sleep state has a state change event.

        Args:
            event: trigger event
        """
        if event.value == habapp_rules.system.SleepState.PRE_SLEEPING.value:
            self.sleep_start()
            if self._config.items.focus is not None and self._config.items.focus.is_on():
                self._config.items.focus.oh_send_command("OFF")
        elif event.value == habapp_rules.system.SleepState.AWAKE.value:
            self.sleep_end()

    def _cb_switch_on(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the switch_on_item has a state change event.

        Args:
            event: trigger event
        """
        if ((self.state == "Auto_HCL" and event.value == "ON") or (isinstance(event.value, int | float) and event.value > 0)) and (target_color := self._get_hcl_color()) is not None:
            self.run.once(1, self._state_observer.send_command, target_color)


class HclElevation(_HclBase):
    """Sun elevation based HCL.

    # Items:
    Number    Elevation                     "Elevation"                     {channel="astro:sun:home:position#elevation"}
    Number    HCL_Color_Elevation           "HCL Color Elevation"
    Switch    HCL_Color_Elevation_manual    "HCL Color Elevation manual"

    # Config
    config = habapp_rules.actors.config.light_hcl.HclElevationConfig(
            items=habapp_rules.actors.config.light_hcl.HclElevationItems(
                    elevation="Elevation",
                    color="HCL_Color_Elevation",
                    manual="HCL_Color_Elevation_manual",
            ),
            parameter=habapp_rules.actors.config.light_hcl.HclElevationParameter(
                    color_map=[(0, 2000), (10, 4000), (30, 5000)]
            )
    )

    # Rule init:
    habapp_rules.actors.light_hcl.HclElevation(config)
    """

    def __init__(self, config: habapp_rules.actors.config.light_hcl.HclElevationConfig) -> None:
        """Init sun elevation based HCL rule.

        Args:
            config: config for HCL rule
        """
        _HclBase.__init__(self, config)
        self._config = config

        self._config.items.elevation.listen_event(self._cb_elevation, HABApp.openhab.events.ItemStateChangedEventFilter())
        self._cb_elevation(None)

    def _get_hcl_color(self) -> int | None:
        """Get HCL color depending on elevation.

        Returns:
            HCL light color
        """
        elevation = self._config.items.elevation.value

        if elevation is None:
            return None

        return_value = 0
        if elevation <= self._config.parameter.color_map[0][0]:
            return_value = self._config.parameter.color_map[0][1]

        elif elevation >= self._config.parameter.color_map[-1][0]:
            return_value = self._config.parameter.color_map[-1][1]

        else:
            for idx, config_itm in enumerate(self._config.parameter.color_map):  # pragma: no cover
                if config_itm[0] <= elevation <= self._config.parameter.color_map[idx + 1][0]:
                    return_value = self._get_interpolated_value(config_itm, self._config.parameter.color_map[idx + 1], elevation)
                    break

        return round(return_value)

    def _cb_elevation(self, _: HABApp.openhab.events.ItemStateChangedEvent | None) -> None:
        """Callback which is called if elevation changed."""
        if self.state == "Auto_HCL" and self._config.items.elevation.value is not None:
            self._state_observer.send_command(self._get_hcl_color())


class HclTime(_HclBase):
    """Time based HCL.

    # Items:
    Number    HCL_Color_Time           "HCL Color Time"
    Switch    HCL_Color_Time_manual    "HCL Color Time manual".

    # Config
    config = habapp_rules.actors.config.light_hcl.HclTimeConfig(
            items=habapp_rules.actors.config.light_hcl.HclTimeItems(
                    color="HCL_Color_Time",
                    manual="HCL_Color_Time_manual",
            ),
            parameter=habapp_rules.actors.config.light_hcl.HclTimeParameter(
                    [(6, 2000), (12, 4000), (20, 3000)],
            )
    )

    # Rule init:
    habapp_rules.actors.light_hcl.HclTime(config)
    """

    def __init__(self, config: habapp_rules.actors.config.light_hcl.HclTimeConfig) -> None:
        """Init time based HCL rule.

        Args:
            config: config for HCL light rule
        """
        _HclBase.__init__(self, config)
        self.run.at(self.run.trigger.interval(None, 300), self._update_color)  # every 5 minutes

    def _one_hour_later(self, current_time: datetime.datetime) -> bool:
        """Check if today the color values will be shifted one hour later in the evening.

        Args:
            current_time: current time

        Returns:
            True if next day is a weekend / holiday day
        """
        if not self._config.parameter.shift_weekend_holiday:
            return False

        if current_time.hour > 12 and (habapp_rules.core.type_of_day.is_holiday(1) or habapp_rules.core.type_of_day.is_weekend(1)):  # noqa: PLR2004
            return True
        return bool(current_time.hour <= 4 and (habapp_rules.core.type_of_day.is_holiday() or habapp_rules.core.type_of_day.is_weekend()))  # noqa: PLR2004

    def _get_hcl_color(self) -> int | None:
        """Get HCL color depending on time.

        Returns:
            HCL light color
        """
        current_time = datetime.datetime.now()

        if self._one_hour_later(current_time):
            current_time -= datetime.timedelta(hours=1)

        if current_time.hour < self._config.parameter.color_map[0][0]:
            start_config = (self._config.parameter.color_map[-1][0] - 24, self._config.parameter.color_map[-1][1])
            end_config = self._config.parameter.color_map[0]

        elif current_time.hour >= self._config.parameter.color_map[-1][0]:
            start_config = self._config.parameter.color_map[-1]
            end_config = (self._config.parameter.color_map[0][0] + 24, self._config.parameter.color_map[0][1])

        else:
            for idx, config_itm in enumerate(self._config.parameter.color_map):  # pragma: no cover
                if config_itm[0] <= current_time.hour < self._config.parameter.color_map[idx + 1][0]:
                    start_config = config_itm
                    end_config = self._config.parameter.color_map[idx + 1]
                    break

        return round(self._get_interpolated_value(start_config, end_config, current_time.hour + current_time.minute / 60))

    def _update_color(self) -> None:
        """Callback which is called every 5 minutes."""
        if self.state == "Auto_HCL":
            self._state_observer.send_command(self._get_hcl_color())



"""Test bathroom light rule."""

import collections
import pathlib
import sys
import time
import unittest
import unittest.mock

import HABApp.rule.rule
from HABApp.core.internals import get_current_context

import habapp_rules.actors.config.energy_save_switch
import habapp_rules.actors.config.light_bathroom
import habapp_rules.actors.config.shading
import habapp_rules.actors.energy_save_switch
import habapp_rules.actors.light_bathroom
import habapp_rules.actors.state_observer
import habapp_rules.core.exceptions
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base
import tests.helper.timer
from habapp_rules.system import PresenceState, SleepState


class TestEnergySaveSwitch(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing energy save switch."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Light_Main")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Light_Main_Ctr")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Light_Main_HCL")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Light_Main_HCL_Lock")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Light_Main_Color")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Light_Mirror")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Presence_state")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Sleep_state")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_State")

        self._config = habapp_rules.actors.config.light_bathroom.BathroomLightConfig(
            items=habapp_rules.actors.config.light_bathroom.BathroomLightItems(
                light_main="Unittest_Light_Main",
                light_main_ctr="Unittest_Light_Main_Ctr",
                light_main_hcl="Unittest_Light_Main_HCL",
                light_main_color="Unittest_Light_Main_Color",
                light_mirror="Unittest_Light_Mirror",
                sleeping_state="Unittest_Sleep_state",
                presence_state="Unittest_Presence_state",
                manual="Unittest_Manual",
                state="Unittest_State",
            )
        )

        self._rule = habapp_rules.actors.light_bathroom.BathroomLight(config=self._config)

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "BathroomLight"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)

        graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(model=tests.helper.graph_machines.FakeModel(), states=self._rule.states, transitions=self._rule.trans, initial=self._rule.state, show_conditions=False)

        graph.get_graph().draw(picture_dir / "BathroomLight.png", format="png", prog="dot")

        for state_name in [state for state in self._get_state_names(self._rule.states) if "init" not in state.lower()]:
            graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(model=tests.helper.graph_machines.FakeModel(), states=self._rule.states, transitions=self._rule.trans, initial=state_name, show_conditions=True)
            graph.get_graph(force_new=True, show_roi=True).draw(picture_dir / f"BathroomLight_{state_name}.png", format="png", prog="dot")

    def test_initial_state(self) -> None:
        """Test initial state."""
        TestCase = collections.namedtuple("TestCase", "manual, sleeping, main_light, mirror_light, expected_state")

        test_cases = [
            # manual OFF | sleeping OFF
            TestCase(manual="OFF", sleeping=False, main_light=0, mirror_light=0, expected_state="Auto_Off"),
            TestCase(manual="OFF", sleeping=False, main_light=0, mirror_light=50, expected_state="Auto_Off"),
            TestCase(manual="OFF", sleeping=False, main_light=100, mirror_light=0, expected_state="Auto_On_MainDay"),
            TestCase(manual="OFF", sleeping=False, main_light=80, mirror_light=100, expected_state="Auto_On_MainAndMirror"),
            # manual OFF | sleeping ON
            TestCase(manual="OFF", sleeping=True, main_light=0, mirror_light=0, expected_state="Auto_Off"),
            TestCase(manual="OFF", sleeping=True, main_light=0, mirror_light=50, expected_state="Auto_Off"),
            TestCase(manual="OFF", sleeping=True, main_light=100, mirror_light=0, expected_state="Auto_On_MainNight"),
            TestCase(manual="OFF", sleeping=True, main_light=80, mirror_light=100, expected_state="Auto_On_MainAndMirror"),
            # manual ON | sleeping OFF
            TestCase(manual="ON", sleeping=False, main_light=0, mirror_light=0, expected_state="Manual"),
            TestCase(manual="ON", sleeping=False, main_light=0, mirror_light=50, expected_state="Manual"),
            TestCase(manual="ON", sleeping=False, main_light=100, mirror_light=0, expected_state="Manual"),
            TestCase(manual="ON", sleeping=False, main_light=80, mirror_light=100, expected_state="Manual"),
            # manual ON | sleeping ON
            TestCase(manual="ON", sleeping=True, main_light=0, mirror_light=0, expected_state="Manual"),
            TestCase(manual="ON", sleeping=True, main_light=0, mirror_light=50, expected_state="Manual"),
            TestCase(manual="ON", sleeping=True, main_light=100, mirror_light=0, expected_state="Manual"),
            TestCase(manual="ON", sleeping=True, main_light=80, mirror_light=100, expected_state="Manual"),
        ]

        with unittest.mock.patch("habapp_rules.actors.light_bathroom.BathroomLight._is_day") as is_day_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    tests.helper.oh_item.set_state("Unittest_Manual", test_case.manual)
                    tests.helper.oh_item.set_state("Unittest_Light_Main", test_case.main_light)
                    tests.helper.oh_item.set_state("Unittest_Light_Mirror", test_case.mirror_light)
                    is_day_mock.return_value = not test_case.sleeping
                    rule = habapp_rules.actors.light_bathroom.BathroomLight(config=self._config)
                    self.assertEqual(test_case.expected_state, rule.state)
                    get_current_context(rule).unload_rule()

    def test_is_day(self) -> None:
        """Test _is_day."""
        TestCase = collections.namedtuple("TestCase", "sleeping_state, sleep_end_time , expected_result")

        test_cases = [
            TestCase(sleeping_state=SleepState.AWAKE.value, sleep_end_time=0, expected_result=True),
            TestCase(sleeping_state=SleepState.AWAKE.value, sleep_end_time=time.time() - 100, expected_result=False),
            TestCase(sleeping_state=SleepState.AWAKE.value, sleep_end_time=time.time() - 10000, expected_result=True),
            TestCase(sleeping_state=SleepState.SLEEPING.value, sleep_end_time=0, expected_result=False),
            TestCase(sleeping_state=SleepState.SLEEPING.value, sleep_end_time=time.time() - 100, expected_result=False),
            TestCase(sleeping_state=SleepState.SLEEPING.value, sleep_end_time=time.time() - 10000, expected_result=False),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_Sleep_state", test_case.sleeping_state)
                self._rule._sleep_end_time = test_case.sleep_end_time
                self.assertEqual(test_case.expected_result, self._rule._is_day())

    def test_is_extended_sleep(self) -> None:
        """Test _is_extended_sleep."""
        TestCase = collections.namedtuple("TestCase", "sleep_end_time, time_now, expected_result")

        self._config.parameter.extended_sleep_time = 10

        test_cases = [
            TestCase(sleep_end_time=100, time_now=200, expected_result=False),
            TestCase(sleep_end_time=100, time_now=100, expected_result=True),
            TestCase(sleep_end_time=100, time_now=109, expected_result=True),
            TestCase(sleep_end_time=100, time_now=110, expected_result=True),
            TestCase(sleep_end_time=100, time_now=111, expected_result=False),
        ]

        with unittest.mock.patch("time.time", return_value=42) as time_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    self._rule._sleep_end_time = test_case.sleep_end_time
                    time_mock.return_value = test_case.time_now
                    self.assertEqual(test_case.expected_result, self._rule._is_extended_sleep())

    def test_mirror_is_on(self) -> None:
        """Test _mirror_is_on."""
        TestCase = collections.namedtuple("TestCase", "mirror_value, expected_result")

        test_cases = [
            TestCase(None, False),
            TestCase(0, False),
            TestCase(1, True),
            TestCase(42, True),
            TestCase(100, True),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_Light_Mirror", test_case.mirror_value)
                self.assertEqual(test_case.expected_result, self._rule._mirror_is_on())

    def test_set_outputs(self) -> None:
        """Test _set_outputs."""
        TestCase = collections.namedtuple("TestCase", "on_via_increase, state, main_initial, main_call, mirror, hcl, color")

        test_cases = [
            # switch_on_via_increase active
            TestCase(on_via_increase=False, state="Manual", main_initial=100, main_call=None, mirror=None, hcl=None, color=None),
            TestCase(on_via_increase=False, state="Auto_Off", main_initial=100, main_call=0, mirror="OFF", hcl=None, color=None),
            TestCase(on_via_increase=False, state="Auto_Off", main_initial=0, main_call=None, mirror="OFF", hcl=None, color=None),
            TestCase(on_via_increase=False, state="Auto_On_MainDay", main_initial=100, main_call=None, mirror=None, hcl="ON", color=None),
            TestCase(on_via_increase=False, state="Auto_On_MainNight", main_initial=100, main_call=40, mirror=None, hcl=None, color=2600),
            TestCase(on_via_increase=False, state="Auto_On_MainAndMirror", main_initial=100, main_call=100, mirror=None, hcl=None, color=4000),
            TestCase(on_via_increase=False, state="Auto_On_MainAndMirror", main_initial=60, main_call=80, mirror=None, hcl=None, color=4000),
            TestCase(on_via_increase=False, state="Auto_On_MainAndMirror", main_initial=90, main_call=90, mirror=None, hcl=None, color=4000),
            # switch_on_via_increase not active
            TestCase(on_via_increase=True, state="Manual", main_initial=100, main_call=None, mirror=None, hcl=None, color=None),
            TestCase(on_via_increase=True, state="Auto_Off", main_initial=100, main_call=0, mirror="OFF", hcl=None, color=None),
            TestCase(on_via_increase=True, state="Auto_Off", main_initial=0, main_call=None, mirror="OFF", hcl=None, color=None),
            TestCase(on_via_increase=True, state="Auto_On_MainDay", main_initial=100, main_call=None, mirror=None, hcl="ON", color=None),
            TestCase(on_via_increase=True, state="Auto_On_MainNight", main_initial=0, main_call=None, mirror=None, hcl=None, color=2600),
            TestCase(on_via_increase=True, state="Auto_On_MainAndMirror", main_initial=100, main_call=100, mirror=None, hcl=None, color=4000),
            TestCase(on_via_increase=True, state="Auto_On_MainAndMirror", main_initial=60, main_call=80, mirror=None, hcl=None, color=4000),
            TestCase(on_via_increase=True, state="Auto_On_MainAndMirror", main_initial=90, main_call=90, mirror=None, hcl=None, color=4000),
        ]

        with unittest.mock.patch("habapp_rules.core.helper.send_if_different") as send_if_different_mock, unittest.mock.patch.object(self._rule, "_light_main_observer") as main_observer_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    send_if_different_mock.reset_mock()
                    main_observer_mock.reset_mock()
                    main_observer_mock.value = test_case.main_initial
                    self._rule.state = test_case.state
                    self._rule._switch_on_via_increase = test_case.on_via_increase
                    self._rule._set_outputs()

                    self.assertEqual(send_if_different_mock.call_count, len([call for call in [test_case.mirror, test_case.hcl, test_case.color] if call is not None]))

                    if test_case.mirror is not None:
                        send_if_different_mock.assert_any_call(self._config.items.light_mirror, test_case.mirror)
                    if test_case.hcl is not None:
                        send_if_different_mock.assert_any_call(self._config.items.light_main_hcl, test_case.hcl)
                    if test_case.color is not None:
                        send_if_different_mock.assert_any_call(self._config.items.light_main_color, test_case.color)

                    if test_case.main_call is not None:
                        main_observer_mock.send_command.assert_called_once_with(test_case.main_call)
                    else:
                        main_observer_mock.send_command.assert_not_called()

                    self.assertFalse(self._rule._switch_on_via_increase)

    def test_set_outputs_sleep_extended(self) -> None:
        """Test _set_outputs for sleep and extended sleep."""
        self._config.parameter.brightness_night = 18
        self._config.parameter.brightness_night_extended = 42
        self._rule.state = "Auto_On_MainNight"

        # normal sleep
        with unittest.mock.patch.object(self._rule, "_is_extended_sleep", return_value=False):
            self._rule._set_outputs()
        tests.helper.oh_item.assert_value("Unittest_Light_Main", 18)

        # extended sleep
        with unittest.mock.patch.object(self._rule, "_is_extended_sleep", return_value=True):
            self._rule._set_outputs()
        tests.helper.oh_item.assert_value("Unittest_Light_Main", 42)

    def test_manual_transitions(self) -> None:
        """Test transitions of state Manual."""
        # set Auto as initial state
        self._rule.to_Auto()

        tests.helper.oh_item.item_state_change_event("Unittest_Manual", "ON")
        self.assertEqual("Manual", self._rule.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Manual", "OFF")
        self.assertEqual("Auto_Off", self._rule.state)

    def test_auto_off_transitions(self) -> None:
        """Test transitions of state Auto_Off."""
        # set Auto as initial state
        self._rule.to_Auto()
        tests.helper.oh_item.assert_value("Unittest_State", "Auto_Off")

        with unittest.mock.patch.object(self._rule, "_is_day", return_value=True) as is_day_mock:
            # main on | day
            tests.helper.oh_item.item_state_change_event("Unittest_Light_Main", 100)
            tests.helper.oh_item.assert_value("Unittest_State", "Auto_On_MainDay")

            # main on | night
            tests.helper.oh_item.item_state_change_event("Unittest_Light_Main", 0)
            tests.helper.oh_item.assert_value("Unittest_State", "Auto_Off")
            is_day_mock.return_value = False
            tests.helper.oh_item.item_state_change_event("Unittest_Light_Main", 100)
            tests.helper.oh_item.assert_value("Unittest_State", "Auto_On_MainNight")

            # main on | night with INCREASE
            tests.helper.oh_item.item_state_change_event("Unittest_Light_Main", 0)
            tests.helper.oh_item.assert_value("Unittest_State", "Auto_Off")
            is_day_mock.return_value = False
            tests.helper.oh_item.item_command_event("Unittest_Light_Main_Ctr", "INCREASE")
            tests.helper.oh_item.assert_value("Unittest_State", "Auto_On_MainNight")
            tests.helper.oh_item.assert_value("Unittest_Light_Main", 0)  # in reality, this will not be 0

            # main on + mirror on | day
            tests.helper.oh_item.item_state_change_event("Unittest_Light_Main", 0)
            tests.helper.oh_item.item_state_change_event("Unittest_Light_Mirror", 100)
            tests.helper.oh_item.assert_value("Unittest_State", "Auto_Off")
            is_day_mock.return_value = True
            tests.helper.oh_item.item_state_change_event("Unittest_Light_Main", 100)
            tests.helper.oh_item.assert_value("Unittest_State", "Auto_On_MainAndMirror")

            # main on + mirror on | night
            tests.helper.oh_item.item_state_change_event("Unittest_Light_Main", 0)
            tests.helper.oh_item.assert_value("Unittest_Light_Mirror", 0)
            tests.helper.oh_item.item_state_change_event("Unittest_Light_Mirror", 100)
            tests.helper.oh_item.assert_value("Unittest_State", "Auto_Off")
            is_day_mock.return_value = True
            tests.helper.oh_item.item_state_change_event("Unittest_Light_Main", 100)
            tests.helper.oh_item.assert_value("Unittest_State", "Auto_On_MainAndMirror")

    def test_sleeping_started(self) -> None:
        """Test sleeping started."""
        self.assertEqual(0, self._rule._sleep_end_time)

        TestCase = collections.namedtuple("TestCase", "sleeping_state, expected_state, expected_sleep_end_time")
        test_cases = [
            TestCase(sleeping_state=SleepState.AWAKE.value, expected_state="Auto_On_MainDay", expected_sleep_end_time=42),
            TestCase(sleeping_state=SleepState.PRE_SLEEPING.value, expected_state="Auto_Off", expected_sleep_end_time=0),
            TestCase(sleeping_state=SleepState.SLEEPING.value, expected_state="Auto_On_MainDay", expected_sleep_end_time=0),
            TestCase(sleeping_state=SleepState.POST_SLEEPING.value, expected_state="Auto_On_MainDay", expected_sleep_end_time=0),
        ]

        with unittest.mock.patch("time.time", return_value=42):
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    self._rule._sleep_end_time = 0
                    self._rule.to_Auto_On_MainDay()
                    tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", test_case.sleeping_state)
                    self.assertEqual(test_case.expected_state, self._rule.state)
                    self.assertEqual(self._rule._sleep_end_time, test_case.expected_sleep_end_time)

    def test_leaving_started(self) -> None:
        """Test leaving started."""
        self.assertEqual(0, self._rule._sleep_end_time)

        TestCase = collections.namedtuple("TestCase", "presence_state, expected_state")
        test_cases = [
            TestCase(presence_state=PresenceState.PRESENCE.value, expected_state="Auto_On_MainDay"),
            TestCase(presence_state=PresenceState.LEAVING.value, expected_state="Auto_Off"),
            TestCase(presence_state=PresenceState.ABSENCE.value, expected_state="Auto_On_MainDay"),
            TestCase(presence_state=PresenceState.LONG_ABSENCE.value, expected_state="Auto_On_MainDay"),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self._rule._sleep_end_time = 0
                self._rule.to_Auto_On_MainDay()
                tests.helper.oh_item.item_state_change_event("Unittest_Presence_state", test_case.presence_state)
                self.assertEqual(test_case.expected_state, self._rule.state)



"""Test irrigation rule."""

import collections
import datetime
import unittest
import unittest.mock

import HABApp

import habapp_rules.actors.config.irrigation
import habapp_rules.actors.irrigation
import habapp_rules.core.exceptions
import tests.helper.oh_item
import tests.helper.test_case_base


class TestIrrigation(tests.helper.test_case_base.TestCaseBase):
    """Tests for Irrigation."""

    def setUp(self) -> None:
        """Set up test cases."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_valve", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_active", "OFF")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_hour", 12)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_minute", 30)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_duration", 5)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_repetitions", 3)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_brake", 10)

        config = habapp_rules.actors.config.irrigation.IrrigationConfig(
            items=habapp_rules.actors.config.irrigation.IrrigationItems(valve="Unittest_valve", active="Unittest_active", hour="Unittest_hour", minute="Unittest_minute", duration="Unittest_duration"), parameter=None
        )

        self._irrigation_min = habapp_rules.actors.irrigation.Irrigation(config)

    def test__init__(self) -> None:
        """Test __init__."""
        self.assertIsNone(self._irrigation_min._config.items.repetitions)
        self.assertIsNone(self._irrigation_min._config.items.brake)

        # init max
        config_max = habapp_rules.actors.config.irrigation.IrrigationConfig(
            items=habapp_rules.actors.config.irrigation.IrrigationItems(
                valve="Unittest_valve", active="Unittest_active", hour="Unittest_hour", minute="Unittest_minute", duration="Unittest_duration", repetitions="Unittest_repetitions", brake="Unittest_brake"
            ),
            parameter=None,
        )

        irrigation_max = habapp_rules.actors.irrigation.Irrigation(config_max)
        self.assertEqual(3, irrigation_max._config.items.repetitions.value)
        self.assertEqual(10, irrigation_max._config.items.brake.value)

    def test_init_with_none(self) -> None:
        """Test __init__ with None values."""
        tests.helper.oh_item.set_state("Unittest_valve", None)
        tests.helper.oh_item.set_state("Unittest_active", None)
        tests.helper.oh_item.set_state("Unittest_hour", None)
        tests.helper.oh_item.set_state("Unittest_minute", None)
        tests.helper.oh_item.set_state("Unittest_duration", None)
        tests.helper.oh_item.set_state("Unittest_repetitions", None)
        tests.helper.oh_item.set_state("Unittest_brake", None)

        config = habapp_rules.actors.config.irrigation.IrrigationConfig(
            items=habapp_rules.actors.config.irrigation.IrrigationItems(
                valve="Unittest_valve", active="Unittest_active", hour="Unittest_hour", minute="Unittest_minute", duration="Unittest_duration", repetitions="Unittest_repetitions", brake="Unittest_brake"
            ),
            parameter=None,
        )

        habapp_rules.actors.irrigation.Irrigation(config)

    def test_get_target_valve_state(self) -> None:
        """Test _get_target_valve_state."""
        # irrigation is not active
        for state in (None, "OFF"):
            tests.helper.oh_item.set_state("Unittest_active", state)
            self.assertFalse(self._irrigation_min._get_target_valve_state())

        # irrigation is active
        tests.helper.oh_item.set_state("Unittest_active", "ON")
        datetime_now = datetime.datetime(2023, 1, 1, 12, 00)
        with unittest.mock.patch("datetime.datetime") as datetime_mock, unittest.mock.patch.object(self._irrigation_min, "_is_in_time_range", return_value=False):
            datetime_mock.now.return_value = datetime_now
            self.assertFalse(self._irrigation_min._get_target_valve_state())
            datetime_mock.combine.assert_called_once_with(date=datetime_now, time=datetime.time(12, 30))
            self._irrigation_min._is_in_time_range.assert_called_once()

        with unittest.mock.patch("datetime.datetime") as datetime_mock, unittest.mock.patch.object(self._irrigation_min, "_is_in_time_range", return_value=True):
            datetime_mock.now.return_value = datetime_now
            self.assertTrue(self._irrigation_min._get_target_valve_state())
            datetime_mock.combine.assert_called_once_with(date=datetime_now, time=datetime.time(12, 30))
            self._irrigation_min._is_in_time_range.assert_called_once()

    def test_get_target_valve_state_with_repetitions(self) -> None:
        """Test _get_target_valve_state with repetitions."""
        config_max = habapp_rules.actors.config.irrigation.IrrigationConfig(
            items=habapp_rules.actors.config.irrigation.IrrigationItems(
                valve="Unittest_valve", active="Unittest_active", hour="Unittest_hour", minute="Unittest_minute", duration="Unittest_duration", repetitions="Unittest_repetitions", brake="Unittest_brake"
            ),
            parameter=None,
        )

        irrigation_max = habapp_rules.actors.irrigation.Irrigation(config_max)
        tests.helper.oh_item.set_state("Unittest_active", "ON")
        tests.helper.oh_item.set_state("Unittest_repetitions", 2)

        # value of hour item is None
        with unittest.mock.patch.object(self._irrigation_min._config.items.hour, "value", None):
            self.assertFalse(self._irrigation_min._get_target_valve_state())

        # value of minute item is None
        with unittest.mock.patch.object(self._irrigation_min._config.items.minute, "value", None):
            self.assertFalse(self._irrigation_min._get_target_valve_state())

        # value of duration item is None
        with unittest.mock.patch.object(self._irrigation_min._config.items.duration, "value", None):
            self.assertFalse(self._irrigation_min._get_target_valve_state())

        # hour, minute and duration are valid
        with unittest.mock.patch.object(irrigation_max, "_is_in_time_range", return_value=False):
            self.assertFalse(irrigation_max._get_target_valve_state())
            self.assertEqual(3, irrigation_max._is_in_time_range.call_count)
            irrigation_max._is_in_time_range.assert_has_calls([
                unittest.mock.call(datetime.time(12, 30), datetime.time(12, 35), unittest.mock.ANY),
                unittest.mock.call(datetime.time(12, 45), datetime.time(12, 50), unittest.mock.ANY),
                unittest.mock.call(datetime.time(13, 0), datetime.time(13, 5), unittest.mock.ANY),
            ])

        with unittest.mock.patch.object(irrigation_max, "_is_in_time_range", side_effect=[False, True]):
            self.assertTrue(irrigation_max._get_target_valve_state())
            self.assertEqual(2, irrigation_max._is_in_time_range.call_count)
            irrigation_max._is_in_time_range.assert_has_calls([
                unittest.mock.call(datetime.time(12, 30), datetime.time(12, 35), unittest.mock.ANY),
                unittest.mock.call(datetime.time(12, 45), datetime.time(12, 50), unittest.mock.ANY),
            ])

    def test_is_in_time_range(self) -> None:
        """Test _is_in_time_range."""
        TestCase = collections.namedtuple("TestCase", "start_time, end_time, time_to_check, expected_result")

        test_cases = [
            TestCase(datetime.time(12, 00), datetime.time(13, 00), datetime.time(12, 30), True),
            TestCase(datetime.time(12, 00), datetime.time(13, 00), datetime.time(14, 30), False),
            TestCase(datetime.time(12, 00), datetime.time(13, 00), datetime.time(13, 00), False),
            TestCase(datetime.time(12, 00), datetime.time(13, 00), datetime.time(12, 00), True),
            TestCase(datetime.time(23, 00), datetime.time(1, 00), datetime.time(23, 0), True),
            TestCase(datetime.time(23, 00), datetime.time(1, 00), datetime.time(23, 59), True),
            TestCase(datetime.time(23, 00), datetime.time(1, 00), datetime.time(0, 0), True),
            TestCase(datetime.time(23, 00), datetime.time(1, 00), datetime.time(0, 30), True),
            TestCase(datetime.time(23, 00), datetime.time(1, 00), datetime.time(1, 0), False),
        ]

        for test_case in test_cases:
            self.assertEqual(test_case.expected_result, self._irrigation_min._is_in_time_range(test_case.start_time, test_case.end_time, test_case.time_to_check))

    def test_cb_set_valve_state(self) -> None:
        """Test _cb_set_valve_state."""
        # called from cyclic call
        with unittest.mock.patch.object(self._irrigation_min, "_get_target_valve_state", return_value=True):
            self._irrigation_min._cb_set_valve_state()
        self.assertEqual("ON", self._irrigation_min._config.items.valve.value)

        # called by event
        with unittest.mock.patch.object(self._irrigation_min, "_get_target_valve_state", return_value=False):
            self._irrigation_min._cb_set_valve_state(HABApp.openhab.events.ItemStateChangedEvent("Unittest_active", "ON", "OFF"))
        self.assertEqual("OFF", self._irrigation_min._config.items.valve.value)

        # same state -> no oh command
        with unittest.mock.patch.object(self._irrigation_min, "_get_target_valve_state", return_value=False), unittest.mock.patch.object(self._irrigation_min._config.items, "valve") as valve_mock:
            valve_mock.is_on.return_value = False
            self._irrigation_min._cb_set_valve_state()
        valve_mock.oh_send_command.assert_not_called()

        # exception at _get_target_valve_stat
        tests.helper.oh_item.set_state("Unittest_valve", "ON")
        with unittest.mock.patch.object(self._irrigation_min, "_get_target_valve_state", side_effect=habapp_rules.core.exceptions.HabAppRulesError("Could not get target state")):
            self._irrigation_min._cb_set_valve_state()
        self.assertEqual("OFF", self._irrigation_min._config.items.valve.value)



"""Tests for SummerWinter Rule."""

import collections
import datetime
import logging
import unittest
import unittest.mock

import HABApp.openhab.definitions.helpers.persistence_data
import HABApp.openhab.items

import habapp_rules.system.config.summer_winter
import habapp_rules.system.summer_winter
import tests.helper.oh_item
import tests.helper.test_case_base


class TestSummerWinter(tests.helper.test_case_base.TestCaseBase):
    """Tests for SummerWinter Rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Temperature", 0)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Summer", "OFF")

        config = habapp_rules.system.config.summer_winter.SummerWinterConfig(items=habapp_rules.system.config.summer_winter.SummerWinterItems(outside_temperature="Unittest_Temperature", summer="Unittest_Summer"))

        self._summer_winter = habapp_rules.system.summer_winter.SummerWinter(config)

    def test_init_with_none(self) -> None:
        """Test __init__ with None values."""
        tests.helper.oh_item.set_state("Unittest_Temperature", None)
        tests.helper.oh_item.set_state("Unittest_Summer", None)

        config = habapp_rules.system.config.summer_winter.SummerWinterConfig(items=habapp_rules.system.config.summer_winter.SummerWinterItems(outside_temperature="Unittest_Temperature", summer="Unittest_Summer"))

        habapp_rules.system.summer_winter.SummerWinter(config)

    def test__get_weighted_mean(self) -> None:
        """Test normal function of wighted_mean."""
        self._summer_winter._config.parameter.persistence_service = "persist_name"
        TestCase = collections.namedtuple("TestCase", "now, expected_day, temperatures, expected_mean")

        test_cases = [
            TestCase(now=datetime.datetime(2050, 1, 1, 17), expected_day=datetime.datetime(2049, 12, 31), temperatures=[[8], [18], [14]], expected_mean=13),
            TestCase(now=datetime.datetime(2050, 1, 1, 23), expected_day=datetime.datetime(2050, 1, 1), temperatures=[[8, 99], [18, 100, 100], [14]], expected_mean=13),
            TestCase(now=datetime.datetime(2050, 1, 1, 22, 59), expected_day=datetime.datetime(2049, 12, 31), temperatures=[[8], [18], [14]], expected_mean=13),
        ]

        with unittest.mock.patch.object(self._summer_winter._config.items, "outside_temperature", spec=HABApp.openhab.items.NumberItem) as outside_temp_mock:
            for test_case in test_cases:
                outside_temp_mock.get_persistence_data.reset_mock()

                # set current time
                self._summer_winter._SummerWinter__now = test_case.now

                # get historical temperatures as HABApp type and set the return to the mock item
                history_temperatures = []
                for temp_list in test_case.temperatures:
                    temp_history = HABApp.openhab.definitions.rest.persistence.ItemHistoryResp(name="some_name", data=[])
                    for idx, temp in enumerate(temp_list):
                        temp_history.data.append(HABApp.openhab.definitions.rest.persistence.DataPoint(time=idx * 123456, state=str(temp)))
                    history_temperatures.append(HABApp.openhab.definitions.helpers.persistence_data.OpenhabPersistenceData.from_resp(temp_history))
                outside_temp_mock.get_persistence_data.side_effect = history_temperatures

                # call weighted mean and check if result is the expected mean temperature
                self.assertTrue(self._summer_winter._SummerWinter__get_weighted_mean(0), test_case.expected_mean)

                # check if call of get_persistence_data was correct
                self.assertEqual(outside_temp_mock.get_persistence_data.call_count, 3)
                outside_temp_mock.get_persistence_data.assert_any_call(persistence="persist_name", start_time=test_case.expected_day + datetime.timedelta(hours=7), end_time=test_case.expected_day + datetime.timedelta(hours=8))
                outside_temp_mock.get_persistence_data.assert_any_call(persistence="persist_name", start_time=test_case.expected_day + datetime.timedelta(hours=14), end_time=test_case.expected_day + datetime.timedelta(hours=15))
                outside_temp_mock.get_persistence_data.assert_any_call(persistence="persist_name", start_time=test_case.expected_day + datetime.timedelta(hours=22), end_time=test_case.expected_day + datetime.timedelta(hours=23))

                # call with days_in_past = 2
                outside_temp_mock.get_persistence_data.side_effect = history_temperatures
                self._summer_winter._SummerWinter__get_weighted_mean(2)

                # check if call of get_persistence_data was correct
                self.assertEqual(outside_temp_mock.get_persistence_data.call_count, 6)
                outside_temp_mock.get_persistence_data.assert_any_call(persistence="persist_name", start_time=test_case.expected_day + datetime.timedelta(days=-2, hours=7), end_time=test_case.expected_day + datetime.timedelta(days=-2, hours=8))
                outside_temp_mock.get_persistence_data.assert_any_call(persistence="persist_name", start_time=test_case.expected_day + datetime.timedelta(days=-2, hours=14), end_time=test_case.expected_day + datetime.timedelta(days=-2, hours=15))
                outside_temp_mock.get_persistence_data.assert_any_call(persistence="persist_name", start_time=test_case.expected_day + datetime.timedelta(days=-2, hours=22), end_time=test_case.expected_day + datetime.timedelta(days=-2, hours=23))

    def test__get_weighted_mean_exception(self) -> None:
        """Test normal function of wighted_mean."""
        with unittest.mock.patch.object(self._summer_winter._config.items, "outside_temperature", spec=HABApp.openhab.items.NumberItem) as outside_temp_mock, self.assertRaises(habapp_rules.system.summer_winter.SummerWinterError) as context:
            outside_temp_mock.get_persistence_data.return_value = HABApp.openhab.definitions.helpers.persistence_data.OpenhabPersistenceData.from_resp(HABApp.openhab.definitions.rest.persistence.ItemHistoryResp(name="some_name", data=[]))
            self._summer_winter._SummerWinter__get_weighted_mean(0)
        self.assertIn("No data for", str(context.exception))

    def test__is_summer(self) -> None:
        """Test if __is_summer method is detecting summer/winter correctly."""
        self._summer_winter._config.parameter.days = 4
        self._summer_winter._SummerWinter__get_weighted_mean = unittest.mock.MagicMock()

        # check if __get_wighted_mean was called correctly
        self._summer_winter._SummerWinter__get_weighted_mean.side_effect = [3, 3.4, 3.6, 4]
        self.assertFalse(self._summer_winter._SummerWinter__is_summer())
        self.assertEqual(self._summer_winter._SummerWinter__get_weighted_mean.call_count, 4)
        self._summer_winter._SummerWinter__get_weighted_mean.assert_any_call(0)
        self._summer_winter._SummerWinter__get_weighted_mean.assert_any_call(1)
        self._summer_winter._SummerWinter__get_weighted_mean.assert_any_call(2)
        self._summer_winter._SummerWinter__get_weighted_mean.assert_any_call(3)

        # check if summer is returned if greater than threshold
        self._summer_winter._SummerWinter__get_weighted_mean.side_effect = [16, 16, 16, 17.1]
        self.assertTrue(self._summer_winter._SummerWinter__is_summer())

        # check if winter is returned if smaller / equal than threshold
        self._summer_winter._SummerWinter__get_weighted_mean.side_effect = [16, 16, 16, 14]
        self.assertFalse(self._summer_winter._SummerWinter__is_summer())

        # check if exceptions are handled correctly (single Exception)
        self._summer_winter._SummerWinter__get_weighted_mean.side_effect = [16, habapp_rules.system.summer_winter.SummerWinterError("not found"), 16.1, 18.0]
        self.assertTrue(self._summer_winter._SummerWinter__is_summer())

        # check if exceptions are handled correctly (single valid value)
        exc = habapp_rules.system.summer_winter.SummerWinterError("not found")
        self._summer_winter._SummerWinter__get_weighted_mean.side_effect = [exc, exc, 16.1, exc]
        self.assertTrue(self._summer_winter._SummerWinter__is_summer())

        # check if exceptions are handled correctly (no value)
        exc = habapp_rules.system.summer_winter.SummerWinterError("not found")
        self._summer_winter._SummerWinter__get_weighted_mean.side_effect = [exc, exc, exc, exc]
        with self.assertRaises(habapp_rules.system.summer_winter.SummerWinterError):
            self._summer_winter._SummerWinter__is_summer()

    def test__is_summer_with_hysteresis(self) -> None:
        """Test summer / winter with hysteresis."""
        TestCase = collections.namedtuple("TestCase", "temperature_values, summer_value, expected_summer")

        test_cases = [
            TestCase([15.74] * 5, False, False),
            TestCase([15.75] * 5, False, False),
            TestCase([15.76] * 5, False, False),
            TestCase([16.24] * 5, False, False),
            TestCase([16.25] * 5, False, True),
            TestCase([16.26] * 5, False, True),
            TestCase([15.74] * 5, True, False),
            TestCase([15.75] * 5, True, True),
            TestCase([15.76] * 5, True, True),
            TestCase([16.24] * 5, True, True),
            TestCase([16.25] * 5, True, True),
            TestCase([16.26] * 5, True, True),
        ]

        self._summer_winter._SummerWinter__get_weighted_mean = unittest.mock.MagicMock()

        for test_case in test_cases:
            self._summer_winter._SummerWinter__get_weighted_mean.side_effect = test_case.temperature_values
            self._summer_winter._hysteresis_switch._on_off_state = test_case.summer_value
            self.assertEqual(test_case.expected_summer, self._summer_winter._SummerWinter__is_summer())

    def test_cb_update_summer(self) -> None:
        """Test correct functionality of summer check callback."""
        with (
            unittest.mock.patch.object(self._summer_winter, "_SummerWinter__is_summer") as is_summer_mock,
            unittest.mock.patch.object(self._summer_winter._config.items, "last_check", spec=HABApp.openhab.items.datetime_item.DatetimeItem) as last_check_mock,
        ):
            # switch from winter to summer
            is_summer_mock.return_value = True
            self._summer_winter._cb_update_summer()
            tests.helper.oh_item.assert_value("Unittest_Summer", "ON")
            self.assertEqual(1, last_check_mock.oh_send_command.call_count)

            # already summer
            is_summer_mock.return_value = True
            with unittest.mock.patch.object(self._summer_winter._config.items, "summer") as summer_item:
                summer_item.value = "ON"
                self._summer_winter._cb_update_summer()
                summer_item.oh_send_command.assert_called_once()
                self.assertEqual(2, last_check_mock.oh_send_command.call_count)

            # switch back to winter
            is_summer_mock.return_value = False
            self._summer_winter._cb_update_summer()
            tests.helper.oh_item.assert_value("Unittest_Summer", "OFF")
            self.assertEqual(3, last_check_mock.oh_send_command.call_count)

            # already winter
            is_summer_mock.return_value = False
            with unittest.mock.patch.object(self._summer_winter._config.items, "summer") as summer_item:
                summer_item.value = "OFF"
                self._summer_winter._cb_update_summer()
                summer_item.oh_send_command.assert_called_once()
                self.assertEqual(4, last_check_mock.oh_send_command.call_count)

            # already winter | no last_check item -> send_command should not be triggered
            is_summer_mock.return_value = False
            with unittest.mock.patch.object(self._summer_winter._config.items, "summer") as summer_item:
                self._summer_winter._config.items.last_check = None
                summer_item.value = "OFF"
                self._summer_winter._cb_update_summer()
                summer_item.oh_send_command.assert_called_once()
                self.assertEqual(4, last_check_mock.oh_send_command.call_count)

        # exception from __is_summer
        with (
            unittest.mock.patch.object(self._summer_winter, "_SummerWinter__is_summer", side_effect=habapp_rules.system.summer_winter.SummerWinterError("No update")),
            unittest.mock.patch.object(self._summer_winter, "_instance_logger", spec=logging.Logger) as logger_mock,
        ):
            self._summer_winter._cb_update_summer()
            logger_mock.exception.assert_called_once()



import datetime

import HABApp.openhab.items
import pydantic

import habapp_rules.core.pydantic_base


class RecurringTaskItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for recurring task."""

    task_active: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item which is set to ON if task is active")
    last_done: HABApp.openhab.items.DatetimeItem = pydantic.Field(..., description="item for date/time when task was last marked as done")


class RecurringTaskParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for recurring task."""

    recurrence_time: datetime.timedelta = pydantic.Field(..., description="recurrence time for task. E.g. if it is set to 10 days, the task will be marked as undone 10 days after it was marked as done")
    fixed_check_time: datetime.time | None = pydantic.Field(None, description="fixed time to check if task must set to done. If set to None, the check time will be calculated based on the recurrence time")

    @pydantic.field_validator("recurrence_time")
    @classmethod
    def cycle_must_be_greater_than_12_hours(cls, v: datetime.timedelta) -> datetime.timedelta:
        """Validate that the cycle is greater than 12 hours.

        Args:
            v: given value

        Returns:
            validated value

        Raises:
            ValueError: if cycle is not greater than 12 hours
        """
        if v < datetime.timedelta(hours=12):
            msg = "Cycle must be greater than 12 hours"
            raise ValueError(msg)
        return v


class RecurringTaskConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for recurring task."""

    items: RecurringTaskItems = pydantic.Field(..., description="items for time task")
    parameter: RecurringTaskParameter = pydantic.Field(..., description="parameter for time task")



"""Tests for Task rules."""

import collections
import datetime

import HABApp.openhab.definitions.helpers.persistence_data
import HABApp.openhab.items

import habapp_rules.system.config.task
import habapp_rules.system.task
import tests.helper.oh_item
import tests.helper.test_case_base


class TestRecurringTaskConfig(tests.helper.test_case_base.TestCaseBase):
    """Tests for RecurringTaskConfig Rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Task", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DatetimeItem, "Unittest_Task_last", None)

        config = habapp_rules.system.config.task.RecurringTaskConfig(
            items=habapp_rules.system.config.task.RecurringTaskItems(task_active="Unittest_Task", last_done="Unittest_Task_last"), parameter=habapp_rules.system.config.task.RecurringTaskParameter(recurrence_time=datetime.timedelta(hours=12))
        )

        self._rule = habapp_rules.system.task.RecurringTask(config)

    def test_init_with_fixed_check_time(self) -> None:
        """Test init with fixed check time."""
        self._rule = habapp_rules.system.task.RecurringTask(
            habapp_rules.system.config.task.RecurringTaskConfig(
                items=habapp_rules.system.config.task.RecurringTaskItems(task_active="Unittest_Task", last_done="Unittest_Task_last"),
                parameter=habapp_rules.system.config.task.RecurringTaskParameter(recurrence_time=datetime.timedelta(hours=12), fixed_check_time=datetime.time(7)),
            )
        )

        self.assertEqual(self._rule._config.parameter.fixed_check_time, datetime.time(7))

    def test__get_check_cycle(self) -> None:
        """Test _get_check_cycle()."""
        TestCase = collections.namedtuple("TestCase", "recurrence_time, expected_result")

        test_cases = [
            TestCase(datetime.timedelta(hours=12), datetime.timedelta(seconds=2160)),
            TestCase(datetime.timedelta(hours=20), datetime.timedelta(hours=1)),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self._rule._config.parameter.recurrence_time = test_case.recurrence_time
                self.assertEqual(self._rule._get_check_cycle(), test_case.expected_result)

    def test_last_done_is_set(self) -> None:
        """Test if last done is set correctly."""
        tests.helper.oh_item.assert_value("Unittest_Task_last", None)

        tests.helper.oh_item.item_state_change_event("Unittest_Task", "ON")
        tests.helper.oh_item.assert_value("Unittest_Task_last", None)

        tests.helper.oh_item.item_state_change_event("Unittest_Task", "OFF")
        self.assertTrue(datetime.datetime.now() - self._rule._config.items.last_done.value < datetime.timedelta(seconds=1))

    def test_check_and_set_task_undone(self) -> None:
        """Test _check_and_set_task_undone."""
        # last done is None
        tests.helper.oh_item.assert_value("Unittest_Task_last", None)
        tests.helper.oh_item.assert_value("Unittest_Task", None)
        self._rule._check_and_set_task_undone()
        tests.helper.oh_item.assert_value("Unittest_Task", "ON")

        # last done is value that should set task to undone
        tests.helper.oh_item.item_state_change_event("Unittest_Task", "OFF")
        tests.helper.oh_item.set_state("Unittest_Task_last", datetime.datetime.now() - datetime.timedelta(days=1))
        self._rule._check_and_set_task_undone()
        tests.helper.oh_item.assert_value("Unittest_Task", "ON")

        # last done is value that should not set task to undone
        tests.helper.oh_item.item_state_change_event("Unittest_Task", "OFF")
        tests.helper.oh_item.set_state("Unittest_Task_last", datetime.datetime.now() - datetime.timedelta(hours=1))
        self._rule._check_and_set_task_undone()
        tests.helper.oh_item.assert_value("Unittest_Task", "OFF")



"""Test config models for sun rules."""

import unittest

import HABApp
import pydantic

import habapp_rules.core.exceptions
import habapp_rules.sensors.config.sun
import tests.helper.oh_item
import tests.helper.test_case_base


class TestTemperatureDifferenceItems(tests.helper.test_case_base.TestCaseBase):
    """Test TemperatureDifferenceItems."""

    def test_validate_temperature_items(self) -> None:
        """Test validate_temperature_items."""
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Output", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Temperature_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Temperature_2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Temperature_3", None)

        # no item is given
        with self.assertRaises(pydantic.ValidationError):
            habapp_rules.sensors.config.sun.TemperatureDifferenceItems(temperatures=[], output="Unittest_Output")

        # single item is given
        with self.assertRaises(pydantic.ValidationError):
            habapp_rules.sensors.config.sun.TemperatureDifferenceItems(temperatures=["Unittest_Temperature_1"], output="Unittest_Output")

        # two items are given
        habapp_rules.sensors.config.sun.TemperatureDifferenceItems(temperatures=["Unittest_Temperature_1", "Unittest_Temperature_2"], output="Unittest_Output")

        # three items are given
        habapp_rules.sensors.config.sun.TemperatureDifferenceItems(temperatures=["Unittest_Temperature_1", "Unittest_Temperature_2", "Unittest_Temperature_3"], output="Unittest_Output")


class TestConfigBase(tests.helper.test_case_base.TestCaseBase):
    """Test ConfigBase."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Brightness", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Output", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Threshold", None)

    def test_validate_threshold(self) -> None:
        """Test validate_threshold."""
        # item NOT given | parameter NOT given
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.sensors.config.sun.BrightnessConfig(items=habapp_rules.sensors.config.sun.BrightnessItems(brightness="Unittest_Brightness", output="Unittest_Output"))

        # item NOT given | parameter given
        habapp_rules.sensors.config.sun.BrightnessConfig(items=habapp_rules.sensors.config.sun.BrightnessItems(brightness="Unittest_Brightness", output="Unittest_Output"), parameter=habapp_rules.sensors.config.sun.BrightnessParameter(threshold=42))

        # item given | parameter NOT given
        habapp_rules.sensors.config.sun.BrightnessConfig(items=habapp_rules.sensors.config.sun.BrightnessItems(brightness="Unittest_Brightness", output="Unittest_Output", threshold="Unittest_Threshold"))

        # item given | parameter given
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.sensors.config.sun.BrightnessConfig(
                items=habapp_rules.sensors.config.sun.BrightnessItems(brightness="Unittest_Brightness", output="Unittest_Output", threshold="Unittest_Threshold"), parameter=habapp_rules.sensors.config.sun.BrightnessParameter(threshold=42)
            )

    def test_threshold_property(self) -> None:
        """Test threshold property."""
        # with parameter
        config = habapp_rules.sensors.config.sun.BrightnessConfig(
            items=habapp_rules.sensors.config.sun.BrightnessItems(brightness="Unittest_Brightness", output="Unittest_Output"), parameter=habapp_rules.sensors.config.sun.BrightnessParameter(threshold=42)
        )
        self.assertEqual(42, config.threshold)

        # with item | value is None
        config = habapp_rules.sensors.config.sun.BrightnessConfig(items=habapp_rules.sensors.config.sun.BrightnessItems(brightness="Unittest_Brightness", output="Unittest_Output", threshold="Unittest_Threshold"))
        self.assertEqual(float("inf"), config.threshold)

        # set value
        tests.helper.oh_item.set_state("Unittest_Threshold", 99)
        self.assertEqual(99, config.threshold)


class TestSunPositionWindow(unittest.TestCase):
    """Tests cases for testing the sun position filter."""

    def test_init(self) -> None:
        """Test __init__."""
        # normal init
        expected_result = habapp_rules.sensors.config.sun.SunPositionWindow(10, 80, 2, 20)
        self.assertEqual(expected_result, habapp_rules.sensors.config.sun.SunPositionWindow(10, 80, 2, 20))

        # init without elevation
        expected_result = habapp_rules.sensors.config.sun.SunPositionWindow(10, 80, 0, 90)
        self.assertEqual(expected_result, habapp_rules.sensors.config.sun.SunPositionWindow(10, 80))

        # init with min > max
        expected_result = habapp_rules.sensors.config.sun.SunPositionWindow(10, 80, 2, 20)
        self.assertEqual(expected_result, habapp_rules.sensors.config.sun.SunPositionWindow(80, 10, 20, 2))



"""Unit-test for filter functions / rules."""

import collections

import HABApp

import habapp_rules.common.config.filter
import habapp_rules.common.filter
import tests.helper.oh_item
import tests.helper.test_case_base


class TestExponentialFilter(tests.helper.test_case_base.TestCaseBase):
    """Tests ExponentialFilter."""

    def setUp(self) -> None:
        """Setup unit-tests."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Raw", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Filtered", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Filtered_2", None)

        config = habapp_rules.common.config.filter.ExponentialFilterConfig(
            items=habapp_rules.common.config.filter.ExponentialFilterItems(raw="Unittest_Raw", filtered="Unittest_Filtered"), parameter=habapp_rules.common.config.filter.ExponentialFilterParameter(tau=10)
        )

        config_increase = habapp_rules.common.config.filter.ExponentialFilterConfig(
            items=habapp_rules.common.config.filter.ExponentialFilterItems(raw="Unittest_Raw", filtered="Unittest_Filtered_2"), parameter=habapp_rules.common.config.filter.ExponentialFilterParameter(tau=100, instant_increase=True)
        )

        self.filter = habapp_rules.common.filter.ExponentialFilter(config)
        self.filter_increase = habapp_rules.common.filter.ExponentialFilter(config_increase)

    def test__init__(self) -> None:
        """Test __init__."""
        self.assertEqual("Unittest_Raw", self.filter._config.items.raw.name)
        self.assertEqual("Unittest_Raw", self.filter_increase._config.items.raw.name)

        self.assertEqual("Unittest_Filtered", self.filter._config.items.filtered.name)
        self.assertEqual("Unittest_Filtered_2", self.filter_increase._config.items.filtered.name)

        self.assertEqual(0.2, self.filter._alpha)
        self.assertEqual(0.2, self.filter_increase._alpha)

        self.assertFalse(self.filter._config.parameter.instant_increase)
        self.assertFalse(self.filter._config.parameter.instant_decrease)

        self.assertTrue(self.filter_increase._config.parameter.instant_increase)
        self.assertFalse(self.filter_increase._config.parameter.instant_decrease)

    def test_cb_cyclic_calculate_and_update_output(self) -> None:
        """Test _cb_cyclic_calculate_and_update_output."""
        TestCase = collections.namedtuple("TestCase", "new_value, previous_value, expected_result")

        test_cases = [
            TestCase(1, 1, 1),
            TestCase(2, 1, 1.2),
            TestCase(2, 0, 0.4),
            TestCase(0, 2, 1.6),
            TestCase(None, None, 999),
            TestCase(None, 42, 999),
            TestCase(42, None, 999),
        ]

        for test_case in test_cases:
            tests.helper.oh_item.set_state("Unittest_Raw", test_case.new_value)
            tests.helper.oh_item.set_state("Unittest_Filtered", 999)  # set some random value
            self.filter._previous_value = test_case.previous_value

            self.filter._cb_cyclic_calculate_and_update_output()

            self.assertEqual(test_case.expected_result, round(self.filter._config.items.filtered.value, 2))

            if test_case.new_value is not None and test_case.previous_value is not None:
                self.assertEqual(test_case.expected_result, round(self.filter._previous_value, 2))
            else:
                self.assertEqual(test_case.previous_value, self.filter._previous_value)

    def test_cb_item_raw(self) -> None:
        """Test _cb_item_raw."""
        TestCase = collections.namedtuple("TestCase", "new_value, previous_value, instant_increase, instant_decrease, expected_value")

        test_cases = [
            TestCase(200, 100, False, False, 100),
            TestCase(200, 100, False, True, 100),
            TestCase(200, 100, True, False, 200),
            TestCase(200, None, False, False, 200),
            TestCase(200, None, False, True, 200),
            TestCase(200, None, True, False, 200),
            TestCase(50, 100, False, False, 100),
            TestCase(50, 100, False, True, 50),
            TestCase(50, 100, True, False, 100),
            TestCase(50, None, False, False, 50),
            TestCase(50, None, False, True, 50),
            TestCase(50, None, True, False, 50),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_Filtered_2", 100)  # set some "random" value
                self.filter_increase._previous_value = test_case.previous_value
                self.filter_increase._config.parameter.instant_increase = test_case.instant_increase
                self.filter_increase._config.parameter.instant_decrease = test_case.instant_decrease

                tests.helper.oh_item.item_state_change_event("Unittest_Raw", test_case.new_value)

                tests.helper.oh_item.assert_value("Unittest_Filtered_2", test_case.expected_value)



"""Test config models for light rules."""

import collections
import unittest.mock

import HABApp
import pydantic

import habapp_rules.actors.config.light
import habapp_rules.core.exceptions
import tests.helper.oh_item
import tests.helper.test_case_base
from habapp_rules.actors.config.light import BrightnessTimeout, FunctionConfig, LightParameter


class TestBrightnessTimeout(unittest.TestCase):
    """Tests for BrightnessTimeout."""

    def test_post_init(self) -> None:
        """Test post init checks."""
        TestCase = collections.namedtuple("TestCase", "value, timeout, valid")

        test_cases = [
            # valid config
            TestCase(100, 1, True),
            TestCase(1, 100, True),
            TestCase(True, 20, True),
            TestCase(False, 0, True),
            TestCase(False, 10, True),
            TestCase(False, 100, True),
            TestCase(0, 100, True),
            # not valid
            TestCase(100, 0, False),
            TestCase(True, 0, False),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                if test_case.valid:
                    brightness_timeout = BrightnessTimeout(test_case.value, test_case.timeout)
                    self.assertEqual(test_case.value, brightness_timeout.brightness)
                    if test_case.value is False:
                        if test_case.timeout:
                            self.assertEqual(test_case.timeout, brightness_timeout.timeout)
                        else:
                            self.assertEqual(0.5, brightness_timeout.timeout)
                else:
                    with self.assertRaises(pydantic.ValidationError):
                        BrightnessTimeout(test_case.value, test_case.timeout)


class TestLightParameter(unittest.TestCase):
    """Tests for LightParameter."""

    def test_post_init(self) -> None:
        """Test check in post init."""
        TestCase = collections.namedtuple("TestCase", "on, pre_off, leaving, pre_sleep, pre_sleep_prevent, valid")

        func_int = FunctionConfig(day=BrightnessTimeout(80, 3), night=BrightnessTimeout(40, 2), sleeping=BrightnessTimeout(20, 1))
        func_int_partial1 = FunctionConfig(day=None, night=BrightnessTimeout(40, 2), sleeping=BrightnessTimeout(20, 1))
        func_int_partial2 = FunctionConfig(day=BrightnessTimeout(80, 3), night=None, sleeping=BrightnessTimeout(20, 1))
        func_int_partial3 = FunctionConfig(day=BrightnessTimeout(80, 3), night=BrightnessTimeout(40, 2), sleeping=None)
        func_bool = FunctionConfig(day=BrightnessTimeout(True, 3), night=BrightnessTimeout(True, 2), sleeping=BrightnessTimeout(True, 1))

        test_cases = [
            TestCase(on=func_bool, pre_off=None, leaving=None, pre_sleep=None, pre_sleep_prevent=None, valid=True),
            TestCase(on=func_bool, pre_off=func_bool, leaving=None, pre_sleep=None, pre_sleep_prevent=None, valid=True),
            TestCase(on=func_bool, pre_off=func_bool, leaving=func_bool, pre_sleep=None, pre_sleep_prevent=None, valid=True),
            TestCase(on=func_bool, pre_off=func_bool, leaving=func_bool, pre_sleep=func_bool, pre_sleep_prevent=None, valid=True),
            TestCase(on=func_int, pre_off=None, leaving=None, pre_sleep=None, pre_sleep_prevent=None, valid=True),
            TestCase(on=func_int, pre_off=func_int, leaving=None, pre_sleep=None, pre_sleep_prevent=None, valid=True),
            TestCase(on=func_int, pre_off=func_int, leaving=func_int, pre_sleep=None, pre_sleep_prevent=None, valid=True),
            TestCase(on=func_int, pre_off=func_int, leaving=func_int, pre_sleep=func_int, pre_sleep_prevent=None, valid=True),
            TestCase(on=None, pre_off=None, leaving=None, pre_sleep=None, pre_sleep_prevent=None, valid=False),
            TestCase(on=None, pre_off=func_bool, leaving=None, pre_sleep=None, pre_sleep_prevent=None, valid=False),
            TestCase(on=None, pre_off=func_bool, leaving=func_bool, pre_sleep=None, pre_sleep_prevent=None, valid=False),
            TestCase(on=None, pre_off=func_bool, leaving=func_bool, pre_sleep=func_bool, pre_sleep_prevent=None, valid=False),
            TestCase(on=None, pre_off=None, leaving=None, pre_sleep=None, pre_sleep_prevent=None, valid=False),
            TestCase(on=None, pre_off=func_int, leaving=None, pre_sleep=None, pre_sleep_prevent=None, valid=False),
            TestCase(on=None, pre_off=func_int, leaving=func_int, pre_sleep=None, pre_sleep_prevent=None, valid=False),
            TestCase(on=None, pre_off=func_int, leaving=func_int, pre_sleep=func_int, pre_sleep_prevent=None, valid=False),
            TestCase(on=func_int_partial1, pre_off=None, leaving=None, pre_sleep=None, pre_sleep_prevent=None, valid=False),
            TestCase(on=func_int_partial2, pre_off=None, leaving=None, pre_sleep=None, pre_sleep_prevent=None, valid=False),
            TestCase(on=func_int_partial3, pre_off=None, leaving=None, pre_sleep=None, pre_sleep_prevent=None, valid=False),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                if test_case.valid:
                    LightParameter(on=test_case.on, pre_off=test_case.pre_off, leaving=test_case.leaving, pre_sleep=test_case.pre_sleep, pre_sleep_prevent=test_case.pre_sleep_prevent)
                else:
                    with self.assertRaises(pydantic.ValidationError):
                        LightParameter(on=test_case.on, pre_off=test_case.pre_off, leaving=test_case.leaving, pre_sleep=test_case.pre_sleep, pre_sleep_prevent=test_case.pre_sleep_prevent)

    def test_sleep_of_pre_sleep(self) -> None:
        """Test if sleep of pre_sleep is set correctly."""
        light_config = LightParameter(
            on=FunctionConfig(day=BrightnessTimeout(True, 3), night=BrightnessTimeout(True, 2), sleeping=BrightnessTimeout(True, 1)), pre_off=None, leaving=None, pre_sleep=FunctionConfig(day=None, night=None, sleeping=BrightnessTimeout(True, 1))
        )

        self.assertIsNone(light_config.pre_sleep.sleeping)


class TestLightConfig(tests.helper.test_case_base.TestCaseBase):
    """Tests for LightConfig."""

    def test_validate_config(self) -> None:
        """Test validate_config."""
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Light", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_CustomState", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Presence_state", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Sleep_state", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Day", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Sleep_prevent", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Door_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Motion", None)

        # ======= min config =======
        habapp_rules.actors.config.light.LightConfig(
            items=habapp_rules.actors.config.light.LightItems(
                light="Unittest_Light",
                manual="Unittest_Manual",
                day="Unittest_Day",
                state="H_CustomState",
            )
        )

        # ======= validate motion =======
        # motion correctly configured
        habapp_rules.actors.config.light.LightConfig(
            items=habapp_rules.actors.config.light.LightItems(light="Unittest_Light", manual="Unittest_Manual", day="Unittest_Day", state="H_CustomState", motion="Unittest_Motion"),
            parameter=habapp_rules.actors.config.light.LightParameter(motion=FunctionConfig(day=None, night=None, sleeping=None)),
        )

        # motion parameter is missing
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.actors.config.light.LightConfig(items=habapp_rules.actors.config.light.LightItems(light="Unittest_Light", manual="Unittest_Manual", day="Unittest_Day", state="H_CustomState", motion="Unittest_Motion"))

        # ======= validate door =======
        # door correctly configured
        habapp_rules.actors.config.light.LightConfig(
            items=habapp_rules.actors.config.light.LightItems(light="Unittest_Light", manual="Unittest_Manual", day="Unittest_Day", state="H_CustomState", doors=["Unittest_Door_1"]),
            parameter=habapp_rules.actors.config.light.LightParameter(door=FunctionConfig(day=None, night=None, sleeping=None)),
        )

        # door parameter is missing
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.actors.config.light.LightConfig(items=habapp_rules.actors.config.light.LightItems(light="Unittest_Light", manual="Unittest_Manual", day="Unittest_Day", state="H_CustomState", doors=["Unittest_Door_1"]))

        # ======= validate sleep =======
        # sleep correctly configured
        habapp_rules.actors.config.light.LightConfig(
            items=habapp_rules.actors.config.light.LightItems(light="Unittest_Light", manual="Unittest_Manual", day="Unittest_Day", state="H_CustomState", sleeping_state="Unittest_Sleep_state"),
            parameter=habapp_rules.actors.config.light.LightParameter(pre_sleep=FunctionConfig(day=None, night=None, sleeping=None)),
        )

        # sleep parameter is missing
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.actors.config.light.LightConfig(
                items=habapp_rules.actors.config.light.LightItems(light="Unittest_Light", manual="Unittest_Manual", day="Unittest_Day", state="H_CustomState", sleeping_state="Unittest_Sleep_state"),
                parameter=habapp_rules.actors.config.light.LightParameter(pre_sleep=None),
            )

        # ======= validate presence =======
        # presence correctly configured
        habapp_rules.actors.config.light.LightConfig(
            items=habapp_rules.actors.config.light.LightItems(light="Unittest_Light", manual="Unittest_Manual", day="Unittest_Day", state="H_CustomState", presence_state="Unittest_Presence_state"),
            parameter=habapp_rules.actors.config.light.LightParameter(leaving=FunctionConfig(day=None, night=None, sleeping=None)),
        )

        # presence parameter is missing
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.actors.config.light.LightConfig(
                items=habapp_rules.actors.config.light.LightItems(light="Unittest_Light", manual="Unittest_Manual", day="Unittest_Day", state="H_CustomState", presence_state="Unittest_Presence_state"),
                parameter=habapp_rules.actors.config.light.LightParameter(leaving=None),
            )

        # ======= validate sleep_prevent =======
        # warning if item and parameter are given
        with unittest.mock.patch.object(habapp_rules.actors.config.light.LOGGER, "warning") as mock_warning:
            habapp_rules.actors.config.light.LightConfig(
                items=habapp_rules.actors.config.light.LightItems(light="Unittest_Light", manual="Unittest_Manual", day="Unittest_Day", state="H_CustomState", pre_sleep_prevent="Unittest_Sleep_prevent"),
                parameter=habapp_rules.actors.config.light.LightParameter(pre_sleep_prevent=unittest.mock.Mock()),
            )
        mock_warning.assert_called_once()



"""Config models for energy_save_switch rules."""

import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class EnergySaveSwitchItems(habapp_rules.core.pydantic_base.ItemBase):
    """Item config for EnergySaveSwitch rule."""

    switch: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="switch item, which will be handled")
    state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="item to store the current state of the state machine")
    manual: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item to switch to manual mode and disable the automatic functions")
    presence_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="presence state set via habapp_rules.presence.Presence")
    sleeping_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="sleeping state set via habapp_rules.system.sleep.Sleep")
    external_request: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item to request ON state from external. This request will overwrite the target state of presence / sleeping.")
    current: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="item which measures the current")


class EnergySaveSwitchParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter config for EnergySaveSwitch rule."""

    max_on_time: int | None = pydantic.Field(None, description="maximum on time in seconds. None means no timeout. If the external request item is ON, the timeout will be extended till the external request is OFF.")
    hand_timeout: int | None = pydantic.Field(None, description="Fallback time from hand to automatic mode in seconds. None means no timeout.")
    current_threshold: float = pydantic.Field(0.030, description="threshold in Ampere.")
    extended_wait_for_current_time: int = pydantic.Field(60, description="Extended time to wait time before switch off the relay in seconds. If current goes above threshold, it will jump back to ON state.", gt=0)


class EnergySaveSwitchConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for EnergySaveSwitch rule."""

    items: EnergySaveSwitchItems = pydantic.Field(..., description="Config items for power switch rule")
    parameter: EnergySaveSwitchParameter = pydantic.Field(EnergySaveSwitchParameter(), description="Config parameter for power switch rule")



import logging

import HABApp
from wakeonlan import send_magic_packet

import habapp_rules.network.config.wol

LOGGER = logging.getLogger(__name__)


class Wol(HABApp.Rule):
    """Rule for wake up a device via WOL.

    Use habapp_rules.network.config.wol.WolConfig to configure this rule.
    """

    def __init__(self, config: habapp_rules.network.config.wol.WolConfig) -> None:
        """Init Rule.

        Args:
            config: config for WOL rule
        """
        HABApp.Rule.__init__(self)
        self._config = config
        self._config.items.trigger_wol.listen_event(self._cb_trigger_wol, HABApp.openhab.events.ItemStateChangedEventFilter())

    def _cb_trigger_wol(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if the trigger_wol item changed.

        Args:
            event: event which triggered this callback

        """
        if event.value == "ON":
            send_magic_packet(self._config.parameter.mac_address)
            LOGGER.info(f"Triggered WOL for '{self._config.parameter.log_name}'")
            self._config.items.trigger_wol.oh_send_command("OFF")



"""Module for filter functions / rules."""

import logging

import HABApp

import habapp_rules.common.config.filter
import habapp_rules.core.logger

LOGGER = logging.getLogger(__name__)


class ExponentialFilter(HABApp.Rule):
    """Rules class to apply a exponential filter to a number value.

    # Items:
    Number    BrightnessValue                       "Brightness Value"                         {channel="..."}
    Number    BrightnessFiltered                    "Brightness filtered"
    Number    BrightnessFilteredInstantIncrease     "Brightness filtered instant increase"

    # Config
    config = habapp_rules.common.config.filter.ExponentialFilterConfig(
            items = habapp_rules.common.config.filter.ExponentialFilterItems(
                    raw = "BrightnessValue",
                    filtered = "BrightnessFiltered"
            ),
            parameter = habapp_rules.common.config.filter.ExponentialFilterParameter(  # filter constant 1 minute
                    tau = 60
           )
    )

    config2 = habapp_rules.common.config.filter.ExponentialFilterConfig(
            items = habapp_rules.common.config.filter.ExponentialFilterItems(
                    raw = "BrightnessValue",
                    filtered = "BrightnessFilteredInstantIncrease"
            ),
            parameter = habapp_rules.common.config.filter.ExponentialFilterParameter(   # filter constant 10 minutes + instant increase
                    tau = 600,
                    instant_increase = True
            )
    )

    # Rule init:
    habapp_rules.common.filter.ExponentialFilter(config)  # filter constant 1 minute
    habapp_rules.common.filter.ExponentialFilter(config2)  # filter constant 10 minutes + instant increase
    """

    def __init__(self, config: habapp_rules.common.config.filter.ExponentialFilterConfig) -> None:
        """Init exponential filter rule.

        Args:
            config: Config for exponential filter
        """
        HABApp.Rule.__init__(self)
        self._config = config

        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self.rule_name)

        self._previous_value = self._config.items.raw.value

        sample_time = self._config.parameter.tau / 5  # fifth part of the filter time constant
        self._alpha = 0.2  # always 0.2 since we always have the fifth part of the filter time constant
        self.run.at(self.run.trigger.interval(None, sample_time), self._cb_cyclic_calculate_and_update_output)

        if self._config.parameter.instant_increase or self._config.parameter.instant_decrease:
            self._config.items.raw.listen_event(self._cb_item_raw, HABApp.openhab.events.ItemStateChangedEventFilter())

        self._instance_logger.debug(f"Successfully created exponential filter for item {self._config.items.raw.name}")

    def _cb_cyclic_calculate_and_update_output(self) -> None:
        """Calculate the new filter output and update the filtered item. This must be called cyclic."""
        new_value = self._config.items.raw.value

        if any(not isinstance(value, int | float) for value in (self._previous_value, new_value)):
            self._instance_logger.warning(f"New or previous value is not a number: new_value: {new_value} | previous_value: {self._previous_value}")
            return

        self._send_output(filtered_value := self._alpha * new_value + (1 - self._alpha) * self._previous_value)
        self._previous_value = filtered_value

    def _cb_item_raw(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is called if the value of the raw item changed.

        Args:
            event: event which triggered this event
        """
        if self._previous_value is None or (self._config.parameter.instant_increase and event.value > self._previous_value) or (self._config.parameter.instant_decrease and event.value < self._previous_value):
            self._send_output(event.value)
            self._previous_value = event.value

    def _send_output(self, new_value: float) -> None:
        """Send output to the OpenHAB item.

        Args:
            new_value: new value which should be sent
        """
        self._config.items.filtered.oh_send_command(new_value)



"""Test DWD rules."""

import collections
import datetime
import pathlib
import sys
import unittest.mock

import HABApp

import habapp_rules.core.exceptions
import habapp_rules.sensors.config.dwd
import habapp_rules.sensors.dwd
import habapp_rules.system
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base


class TestDwdItems(tests.helper.test_case_base.TestCaseBase):
    """Tests for DwdItems."""

    def setUp(self) -> None:
        """Setup tests."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "I26_99_warning_1_description", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "I26_99_warning_1_type", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "I26_99_warning_1_severity", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DatetimeItem, "I26_99_warning_1_start_time", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DatetimeItem, "I26_99_warning_1_end_time", None)

        self._test_dataclass = habapp_rules.sensors.dwd.DwdItems.from_prefix("I26_99_warning_1")

    def test_severity_as_int(self) -> None:
        """Test severity_as_int."""
        TestCase = collections.namedtuple("TestCase", "str_value, expected_int")

        test_cases = [
            TestCase("NULL", 0),
            TestCase("Minor", 1),
            TestCase("Moderate", 2),
            TestCase("Severe", 3),
            TestCase("Extreme", 4),
            TestCase("UNKNOWN", 0),
            TestCase(None, 0),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("I26_99_warning_1_severity", test_case.str_value)
                self.assertEqual(test_case.expected_int, self._test_dataclass.severity_as_int)


class TestDwdWindAlarm(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests for DwdWindAlarm."""

    def setUp(self) -> None:
        """Setup tests."""
        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Wind_Alarm_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Wind_Alarm_1_state", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Wind_Alarm_2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual_2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Wind_Alarm_2_state", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Hand_Timeout", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "I26_99_warning_1_description", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "I26_99_warning_1_type", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "I26_99_warning_1_severity", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DatetimeItem, "I26_99_warning_1_start_time", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DatetimeItem, "I26_99_warning_1_end_time", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "I26_99_warning_2_description", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "I26_99_warning_2_type", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "I26_99_warning_2_severity", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DatetimeItem, "I26_99_warning_2_start_time", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DatetimeItem, "I26_99_warning_2_end_time", None)

        config_1 = habapp_rules.sensors.config.dwd.WindAlarmConfig(
            items=habapp_rules.sensors.config.dwd.WindAlarmItems(wind_alarm="Unittest_Wind_Alarm_1", manual="Unittest_Manual_1", state="H_Unittest_Wind_Alarm_1_state"),
            parameter=habapp_rules.sensors.config.dwd.WindAlarmParameter(hand_timeout=12 * 3600, number_dwd_objects=2),
        )

        config_2 = habapp_rules.sensors.config.dwd.WindAlarmConfig(
            items=habapp_rules.sensors.config.dwd.WindAlarmItems(wind_alarm="Unittest_Wind_Alarm_2", manual="Unittest_Manual_2", hand_timeout="Unittest_Hand_Timeout", state="Unittest_Wind_Alarm_2_state"),
            parameter=habapp_rules.sensors.config.dwd.WindAlarmParameter(number_dwd_objects=2),
        )

        self._wind_alarm_rule_1 = habapp_rules.sensors.dwd.DwdWindAlarm(config_1)
        self._wind_alarm_rule_2 = habapp_rules.sensors.dwd.DwdWindAlarm(config_2)

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "DWD_WindAlarm"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)

        graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(
            model=tests.helper.graph_machines.FakeModel(), states=self._wind_alarm_rule_1.states, transitions=self._wind_alarm_rule_1.trans, initial=self._wind_alarm_rule_1.state, show_conditions=True
        )

        graph.get_graph().draw(picture_dir / "DWD_Wind_Alarm.png", format="png", prog="dot")

    def test_set_timeouts(self) -> None:
        """Test _set_timeouts."""
        self.assertEqual(12 * 3600, self._wind_alarm_rule_1.state_machine.get_state("Hand").timeout)
        self.assertEqual(24 * 3600, self._wind_alarm_rule_2.state_machine.get_state("Hand").timeout)

        tests.helper.oh_item.item_state_change_event("Unittest_Hand_Timeout", 2000)
        self.assertEqual(2000, self._wind_alarm_rule_2.state_machine.get_state("Hand").timeout)

    def test_get_initial_state(self) -> None:
        """Test _get_initial_state."""
        TestCase = collections.namedtuple("TestCase", "manual, wind_alarm_active, expected_state")

        test_cases = [
            TestCase("OFF", False, "Auto_Off"),
            TestCase("OFF", True, "Auto_On"),
            TestCase("ON", False, "Manual"),
            TestCase("ON", True, "Manual"),
        ]

        with unittest.mock.patch.object(self._wind_alarm_rule_1, "_wind_alarm_active") as wind_alarm_active_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    tests.helper.oh_item.set_state("Unittest_Manual_1", test_case.manual)
                    wind_alarm_active_mock.return_value = test_case.wind_alarm_active

                    self.assertEqual(test_case.expected_state, self._wind_alarm_rule_1._get_initial_state())

    def test_manual(self) -> None:
        """Test manual."""
        # from Auto
        self.assertEqual("Auto_Off", self._wind_alarm_rule_1.state)
        self.assertEqual("Auto_Off", self._wind_alarm_rule_2.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Manual_1", "ON")
        tests.helper.oh_item.item_state_change_event("Unittest_Manual_2", "ON")
        self.assertEqual("Manual", self._wind_alarm_rule_1.state)
        self.assertEqual("Manual", self._wind_alarm_rule_2.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Manual_1", "OFF")
        tests.helper.oh_item.item_state_change_event("Unittest_Manual_2", "OFF")
        self.assertEqual("Auto_Off", self._wind_alarm_rule_1.state)
        self.assertEqual("Auto_Off", self._wind_alarm_rule_2.state)

        # from Hand
        tests.helper.oh_item.item_state_change_event("Unittest_Wind_Alarm_1", "ON")
        tests.helper.oh_item.item_state_change_event("Unittest_Wind_Alarm_2", "ON")
        self.assertEqual("Hand", self._wind_alarm_rule_1.state)
        self.assertEqual("Hand", self._wind_alarm_rule_2.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Manual_1", "ON")
        tests.helper.oh_item.item_state_change_event("Unittest_Manual_2", "ON")
        self.assertEqual("Manual", self._wind_alarm_rule_1.state)
        self.assertEqual("Manual", self._wind_alarm_rule_2.state)

    def test_wind_alarm_active(self) -> None:
        """Test _wind_alarm_active."""
        TestCase = collections.namedtuple("TestCase", "type_1, description_1, severity_1 start_time_1, end_time_1, type_2, description_2, severity_2 start_time_2, end_time_2, expected_result")

        now = datetime.datetime.now()
        start_active = now + datetime.timedelta(hours=-1)
        end_active = now + datetime.timedelta(hours=1)
        start_not_active = now + datetime.timedelta(hours=1)
        end_not_active = now + datetime.timedelta(hours=-2)

        test_cases = [
            TestCase(None, None, None, None, None, None, None, None, None, None, False),
            TestCase("FROST", "Frost is appearing at 100 km/h", "Minor", start_not_active, end_not_active, "SUN", "SUN is appearing at 100 km/h", "Minor", start_not_active, end_not_active, False),
            TestCase("FROST", "Frost is appearing at 100 km/h", "Minor", start_not_active, end_not_active, "SUN", "SUN is appearing at 100 km/h", "Minor", start_active, end_active, False),
            TestCase("FROST", "Frost is appearing at 100 km/h", "Minor", start_not_active, end_not_active, "SUN", "SUN is appearing at 100 km/h", "Moderate", start_active, end_active, False),
            TestCase("FROST", "Frost is appearing at 100 km/h", "Minor", start_not_active, end_not_active, "SUN", "Wind speed above 100 km/h", "Moderate", start_active, end_active, False),
            TestCase("FROST", "Frost is appearing at 100 km/h", "Minor", start_not_active, end_not_active, "WIND", "Wind speed above 100 km/h", "Moderate", start_active, end_active, True),
            TestCase("FROST", "Frost is appearing at 100 km/h", "Minor", start_not_active, end_not_active, "WIND", "Wind speed very high", "Moderate", start_active, end_active, False),
            TestCase("FROST", "Frost is appearing at 100 km/h", "Minor", start_not_active, end_not_active, "WIND", "Wind speed above 100 km/h", "Minor", start_active, end_active, False),
            TestCase("FROST", "Frost is appearing at 100 km/h", "Minor", start_not_active, end_not_active, "WIND", "Wind speed above 10 km/h", "Extreme", start_active, end_active, False),
            TestCase("FROST", "Frost is appearing at 100 km/h", "Minor", start_not_active, end_not_active, "WIND", "Wind speed between 5 km/h and 100 km/h", "Moderate", start_active, end_active, True),
            TestCase("FROST", "Frost is appearing at 100 km/h", "Minor", start_not_active, end_not_active, "WIND", "Wind speed between 5 km/h and 100 km/h", "Moderate", start_not_active, end_not_active, False),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("I26_99_warning_1_description", test_case.description_1)
                tests.helper.oh_item.set_state("I26_99_warning_1_type", test_case.type_1)
                tests.helper.oh_item.set_state("I26_99_warning_1_severity", test_case.severity_1)
                tests.helper.oh_item.set_state("I26_99_warning_1_start_time", test_case.start_time_1)
                tests.helper.oh_item.set_state("I26_99_warning_1_end_time", test_case.end_time_1)

                tests.helper.oh_item.set_state("I26_99_warning_2_description", test_case.description_2)
                tests.helper.oh_item.set_state("I26_99_warning_2_type", test_case.type_2)
                tests.helper.oh_item.set_state("I26_99_warning_2_severity", test_case.severity_2)
                tests.helper.oh_item.set_state("I26_99_warning_2_start_time", test_case.start_time_2)
                tests.helper.oh_item.set_state("I26_99_warning_2_end_time", test_case.end_time_2)

                self.assertEqual(test_case.expected_result, self._wind_alarm_rule_1._wind_alarm_active())

    def test_cyclic_check(self) -> None:
        """Test _cyclic_check."""
        # Manual / Hand should not trigger any test
        with unittest.mock.patch.object(self._wind_alarm_rule_1, "_wind_alarm_active") as check_wind_alarm_mock:
            for state in ("Manual", "Hand"):
                self._wind_alarm_rule_1.state = state
                self._wind_alarm_rule_1._cb_cyclic_check()
        check_wind_alarm_mock.assert_not_called()

        # Auto will trigger check and send if needed
        TestCase = collections.namedtuple("TestCase", "initial_state, wind_alarm_active, expected_state")

        test_cases = [
            TestCase("Auto_Off", False, "Auto_Off"),
            TestCase("Auto_Off", True, "Auto_On"),
            TestCase("Auto_On", True, "Auto_On"),
            TestCase("Auto_On", False, "Auto_Off"),
        ]

        with unittest.mock.patch.object(self._wind_alarm_rule_1, "_wind_alarm_active") as check_wind_alarm_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    self._wind_alarm_rule_1.state = test_case.initial_state
                    check_wind_alarm_mock.return_value = test_case.wind_alarm_active

                    self._wind_alarm_rule_1._cb_cyclic_check()

                    self.assertEqual(test_case.expected_state, self._wind_alarm_rule_1.state)
                    tests.helper.oh_item.assert_value("Unittest_Wind_Alarm_1", "ON" if test_case.expected_state == "Auto_On" else "OFF")



"""Config models for monthly energy report."""

import datetime

import HABApp
import multi_notifier.connectors.connector_mail
import pydantic

import habapp_rules.core.pydantic_base
import habapp_rules.energy.helper


class EnergyShare(pydantic.BaseModel):
    """Dataclass for defining energy share objects."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    energy_item: HABApp.openhab.items.NumberItem | list[HABApp.openhab.items.NumberItem]
    chart_name: str
    monthly_power: float = 0.0

    def __init__(self, energy_item: str | HABApp.openhab.items.NumberItem | list[HABApp.openhab.items.NumberItem] | list[str], chart_name: str, monthly_power: float = 0.0) -> None:
        """Init energy share object without keywords.

        Args:
            energy_item: name or item of energy
            chart_name: name which will be shown in the chart
            monthly_power: monthly power of this energy share. This will be set by the energy share rule.
        """
        super().__init__(energy_item=energy_item, chart_name=chart_name, monthly_power=monthly_power)

    @staticmethod
    def _get_number_item_by_name(name: str) -> HABApp.openhab.items.NumberItem:
        try:
            return HABApp.openhab.items.NumberItem.get_item(name)
        except HABApp.core.errors.ItemNotFoundException as exc:
            msg = f"Could not find any item for given name '{name}'"
            raise ValueError(msg) from exc

    @pydantic.field_validator("energy_item", mode="before")
    @classmethod
    def check_oh_item(cls, data: str | HABApp.openhab.items.NumberItem) -> HABApp.openhab.items.NumberItem | list[HABApp.openhab.items.NumberItem]:
        """Check if given item is an OpenHAB item or try to get it from OpenHAB.

        Args:
            data: configuration for energy item

        Returns:
            energy item

        Raises:
            ValueError: if item could not be found
        """
        if isinstance(data, HABApp.openhab.items.NumberItem):
            return data
        if isinstance(data, list):
            if all(isinstance(itm, HABApp.openhab.items.NumberItem) for itm in data):
                return data
            return [cls._get_number_item_by_name(itm) for itm in data]
        return cls._get_number_item_by_name(data)

    def get_energy_since(self, start_time: datetime.datetime) -> float:
        """Get energy since start time.

        Args:
            start_time: start time to search for the interested value

        Returns:
            energy since start time
        """
        if isinstance(self.energy_item, list):
            return sum(itm.value - habapp_rules.energy.helper.get_historic_value(itm, start_time) for itm in self.energy_item)
        return self.energy_item.value - habapp_rules.energy.helper.get_historic_value(self.energy_item, start_time)

    @property
    def get_items_as_list(self) -> list[HABApp.openhab.items.NumberItem]:
        """Get energy item(s) as list.

        Returns:
            All energy items
        """
        if isinstance(self.energy_item, list):
            return self.energy_item
        return [self.energy_item]


class MonthlyReportItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for monthly report."""

    energy_sum: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="item which holds the total energy consumption")


class MonthlyReportParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for monthly report."""

    known_energy_shares: list[EnergyShare] = pydantic.Field([], description="list of EnergyShare objects which hold the known energy shares. E.g. energy for lights or ventilation")
    persistence_group_name: str | None = pydantic.Field(None, description="OpenHAB group name which holds all items which are persisted. If the group name is given it will be checked if all energy items are in the group")
    config_mail: multi_notifier.connectors.connector_mail.MailConfig = pydantic.Field(..., description="config for sending mails")
    recipients: list[str] = pydantic.Field(..., description="list of recipients who get the mail")
    debug: bool = pydantic.Field(default=False, description="if debug mode is active")


class MonthlyReportConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for monthly report."""

    items: MonthlyReportItems = pydantic.Field(..., description="Items for monthly report")
    parameter: MonthlyReportParameter = pydantic.Field(..., description="Parameter for monthly report")



"""Test config models of dwd rules."""

import HABApp

import habapp_rules.core.exceptions
import habapp_rules.sensors.config.dwd
import tests.helper.oh_item
import tests.helper.test_case_base


class TestWindAlarmConfig(tests.helper.test_case_base.TestCaseBase):
    """Test WindAlarmConfig."""

    def test_check_hand_timeout(self) -> None:
        """Test check_hand_timeout."""
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Wind_Alarm", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Hand_Timeout", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Wind_Alarm_state", None)

        # no timeout is given
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.sensors.config.dwd.WindAlarmConfig(items=habapp_rules.sensors.config.dwd.WindAlarmItems(wind_alarm="Unittest_Wind_Alarm", manual="Unittest_Manual", state="H_Unittest_Wind_Alarm_state"))

        # only timeout item is given
        habapp_rules.sensors.config.dwd.WindAlarmConfig(items=habapp_rules.sensors.config.dwd.WindAlarmItems(wind_alarm="Unittest_Wind_Alarm", manual="Unittest_Manual", hand_timeout="Unittest_Hand_Timeout", state="H_Unittest_Wind_Alarm_state"))

        # only timeout parameter is given
        habapp_rules.sensors.config.dwd.WindAlarmConfig(
            items=habapp_rules.sensors.config.dwd.WindAlarmItems(wind_alarm="Unittest_Wind_Alarm", manual="Unittest_Manual", state="H_Unittest_Wind_Alarm_state"), parameter=habapp_rules.sensors.config.dwd.WindAlarmParameter(hand_timeout=12 * 3600)
        )

        # timeout parameter and item are given
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.sensors.config.dwd.WindAlarmConfig(
                items=habapp_rules.sensors.config.dwd.WindAlarmItems(wind_alarm="Unittest_Wind_Alarm", manual="Unittest_Manual", hand_timeout="Unittest_Hand_Timeout", state="H_Unittest_Wind_Alarm_state"),
                parameter=habapp_rules.sensors.config.dwd.WindAlarmParameter(hand_timeout=12 * 3600),
            )



"""Unittest for TimeoutList."""

import unittest
import unittest.mock

import habapp_rules.core.timeout_list


class TestValueWithTimeout(unittest.TestCase):
    """Tests for ValueWithTimeout."""

    def test_less_than(self) -> None:
        """Test less than of ValueWithTimeout."""
        self.assertTrue(habapp_rules.core.timeout_list.ValueWithTimeout(42, 10, 10) < habapp_rules.core.timeout_list.ValueWithTimeout(80, 10, 10))
        self.assertFalse(habapp_rules.core.timeout_list.ValueWithTimeout(80, 10, 10) < habapp_rules.core.timeout_list.ValueWithTimeout(42, 10, 10))
        self.assertFalse(habapp_rules.core.timeout_list.ValueWithTimeout(42, 10, 10) < habapp_rules.core.timeout_list.ValueWithTimeout(42, 10, 10))
        with self.assertRaises(TypeError):
            self.assertEqual(NotImplemented, habapp_rules.core.timeout_list.ValueWithTimeout(42, 10, 10) < 42)

    def test_equal(self) -> None:
        """Test equal of ValueWithTimeout."""
        self.assertTrue(habapp_rules.core.timeout_list.ValueWithTimeout(42, 10, 10) == habapp_rules.core.timeout_list.ValueWithTimeout(42, 10, 10))
        self.assertFalse(habapp_rules.core.timeout_list.ValueWithTimeout(42, 10, 10) == habapp_rules.core.timeout_list.ValueWithTimeout(80, 10, 10))
        self.assertFalse(habapp_rules.core.timeout_list.ValueWithTimeout(42, 10, 10) == 42)

    def test_hash(self) -> None:
        """Test hash of ValueWithTimeout."""
        self.assertEqual(hash(habapp_rules.core.timeout_list.ValueWithTimeout(42, 10, 10)), hash(habapp_rules.core.timeout_list.ValueWithTimeout(42, 10, 10)))


class TestTimeoutList(unittest.TestCase):
    """Tests for TimeoutList."""

    def setUp(self) -> None:
        """Setup Tests."""
        self.normal_list = []
        self.timeout_list = habapp_rules.core.timeout_list.TimeoutList()

    def test_repr(self) -> None:
        """Test repr of TimeoutList."""
        self.assertEqual(str(self.normal_list), str(self.timeout_list))
        for add_value in [None, "some_string", 14, 42.2]:
            self.normal_list.append(add_value)
            self.timeout_list.append(add_value, 100)
            self.assertEqual(str(self.normal_list), str(self.timeout_list))

    def test_bool(self) -> None:
        """Test bool of TimeoutList."""
        with unittest.mock.patch.object(self.timeout_list, "_TimeoutList__remove_old_items") as remove_mock:
            self.assertFalse(bool(self.timeout_list))
            remove_mock.assert_called_once()

            remove_mock.reset_mock()
            self.timeout_list.append(42, 10)
            self.assertTrue(bool(self.timeout_list))
            remove_mock.assert_called_once()

    def test_contains(self) -> None:
        """Test contains of TimeoutList."""
        with unittest.mock.patch.object(self.timeout_list, "_TimeoutList__remove_old_items") as remove_mock:
            # empty list
            self.assertFalse(42 in self.timeout_list)
            self.assertFalse("test" in self.timeout_list)
            self.assertEqual(2, remove_mock.call_count)

            # add 42
            remove_mock.reset_mock()
            self.timeout_list.append(42, 10)
            self.assertTrue(42 in self.timeout_list)
            self.assertFalse("test" in self.timeout_list)
            self.assertEqual(2, remove_mock.call_count)

            # add test string
            remove_mock.reset_mock()
            self.timeout_list.append("test", 10)
            self.assertTrue(42 in self.timeout_list)
            self.assertTrue("test" in self.timeout_list)
            self.assertEqual(2, remove_mock.call_count)

    def test_get_item(self) -> None:
        """Test getting item from TimeoutList."""
        # empty list
        with self.assertRaises(IndexError):
            self.assertIsNone(self.timeout_list[0])

        with self.assertRaises(IndexError):
            self.assertIsNone(self.timeout_list[1])

        # list with two elements
        self.timeout_list.append("test", 10)
        self.timeout_list.append(42, 10)

        self.assertEqual("test", self.timeout_list[0])
        self.assertEqual(42, self.timeout_list[1])

        with self.assertRaises(IndexError):
            self.assertIsNone(self.timeout_list[2])

    def test_equal(self) -> None:
        """Test equal of TimeoutList."""
        with unittest.mock.patch.object(self.timeout_list, "_TimeoutList__remove_old_items") as remove_old_mock:
            self.assertEqual(self.timeout_list, habapp_rules.core.timeout_list.TimeoutList())
            self.assertEqual(1, remove_old_mock.call_count)

            self.assertEqual(self.timeout_list, [])
            self.assertEqual(2, remove_old_mock.call_count)

            self.timeout_list.append(42, 10)
            self.assertEqual(self.timeout_list, [42])
            self.assertEqual(3, remove_old_mock.call_count)

            self.assertNotEqual(self.timeout_list, "")
            self.assertEqual(3, remove_old_mock.call_count)

    def test_not_equal(self) -> None:
        """Test not equal of TimeoutList."""
        self.assertFalse(self.timeout_list != [])

        self.timeout_list.append(42, 10)
        self.assertTrue(self.timeout_list != [])
        self.assertTrue(self.timeout_list != [80])
        self.assertFalse(self.timeout_list != [42])

    def test_iter(self) -> None:
        """Test iter of TimeoutList."""
        self.assertEqual([], list(iter(self.timeout_list)))

        self.timeout_list.append(42, 10)
        self.assertEqual([42], list(iter(self.timeout_list)))

        self.timeout_list.append("test", 10)
        self.assertEqual([42, "test"], list(iter(self.timeout_list)))

    def test_remove(self) -> None:
        """Test remove of TimeoutList."""
        with self.assertRaises(ValueError) as context:
            self.timeout_list.remove(42)
        self.assertEqual("TimeoutList.remove(x): x not in list", str(context.exception))

        self.timeout_list.append(42, 10)
        self.timeout_list.append("test", 10)
        self.assertIsNone(self.timeout_list.remove(42))

        self.assertEqual(self.timeout_list, ["test"])

    def test_pop(self) -> None:
        """Test pop of TimeoutList."""
        with self.assertRaises(IndexError):
            self.timeout_list.pop(0)

        self.timeout_list.append(42, 10)

        with self.assertRaises(IndexError):
            self.timeout_list.pop(1)

        with self.assertRaises(TypeError):
            self.timeout_list.pop("first")

        self.assertEqual(42, self.timeout_list.pop(0))
        self.assertEqual([], self.timeout_list)

        self.timeout_list.append(42, 10)
        self.timeout_list.append("test", 10)
        self.assertEqual("test", self.timeout_list.pop(1))
        self.assertEqual([42], self.timeout_list)

    def test_hash(self) -> None:
        """Test hash method."""
        self.assertEqual(5740354900026072187, hash(self.timeout_list))
        self.assertEqual(hash(()), hash(self.timeout_list))

        self.timeout_list.append(2, 2)
        self.assertEqual(hash((2,)), hash(self.timeout_list))

        self.timeout_list.append(42, 100)
        self.assertEqual(hash((2, 42)), hash(self.timeout_list))



import datetime
import unittest

from habapp_rules.system.config.task import RecurringTaskParameter


class TestRecurringTaskParameter(unittest.TestCase):
    """Test RecurringTaskParameter."""

    def test_validation(self) -> None:
        """Test validation of RecurringTaskParameter."""
        # valid (time greater than 12 hours)
        RecurringTaskParameter(recurrence_time=datetime.timedelta(hours=13))

        # valid (time equal 12 hours)
        RecurringTaskParameter(recurrence_time=datetime.timedelta(hours=12))

        # invalid (time less than 12 hours)
        with self.assertRaises(ValueError):
            RecurringTaskParameter(recurrence_time=datetime.timedelta(hours=11))



"""Test config models for shading rules."""

import collections
import unittest
from itertools import starmap

import HABApp
import pydantic

import habapp_rules.actors.config.shading
import habapp_rules.core.exceptions
import tests.helper.oh_item
import tests.helper.test_case_base


class TestShadingConfig(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing ShadingConfig."""

    def tests_validate_model(self) -> None:
        """Test validate_model."""
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.RollershutterItem, "Unittest_Shading", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Manual", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Shading_state", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Summer", None)

        # parameter NOT given | item summer NOT given
        habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(shading_position="Unittest_Shading", manual="Unittest_Manual", state="H_Unittest_Shading_state"), parameter=habapp_rules.actors.config.shading.ShadingParameter()
        )

        # parameter NOT given | item summer given
        habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(shading_position="Unittest_Shading", manual="Unittest_Manual", state="H_Unittest_Shading_state", summer="Unittest_Summer"),
            parameter=habapp_rules.actors.config.shading.ShadingParameter(),
        )

        # parameter given | item summer NOT given
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.actors.config.shading.ShadingConfig(
                items=habapp_rules.actors.config.shading.ShadingItems(shading_position="Unittest_Shading", manual="Unittest_Manual", state="H_Unittest_Shading_state"),
                parameter=habapp_rules.actors.config.shading.ShadingParameter(pos_night_close_summer=habapp_rules.actors.config.shading.ShadingPosition(42, 80)),
            )

        # parameter given | item summer given
        habapp_rules.actors.config.shading.ShadingConfig(
            items=habapp_rules.actors.config.shading.ShadingItems(shading_position="Unittest_Shading", manual="Unittest_Manual", state="H_Unittest_Shading_state", summer="Unittest_Summer"),
            parameter=habapp_rules.actors.config.shading.ShadingParameter(pos_night_close_summer=habapp_rules.actors.config.shading.ShadingPosition(42, 80)),
        )


class TestSlatValueParameter(unittest.TestCase):
    """Test slat value parameter."""

    def test__check_and_sort_characteristic(self) -> None:
        """Test __check_and_sort_characteristic."""
        TestCase = collections.namedtuple("TestCase", "input, expected_output, raises")

        test_cases = [
            TestCase([(0, 100), (10, 50)], [(0, 100), (10, 50)], False),
            TestCase([(10, 50), (0, 100)], [(0, 100), (10, 50)], False),
            TestCase([(0, 100), (10, 50), (20, 50)], [(0, 100), (10, 50), (20, 50)], False),
            TestCase([(10, 50), (0, 100), (20, 50)], [(0, 100), (10, 50), (20, 50)], False),
            TestCase([(10, 50), (20, 50), (0, 100)], [(0, 100), (10, 50), (20, 50)], False),
            TestCase([(0, 50), (0, 40)], None, True),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                input_conf = list(starmap(habapp_rules.actors.config.shading.ElevationSlatMapping, test_case.input))
                output = list(starmap(habapp_rules.actors.config.shading.ElevationSlatMapping, test_case.expected_output)) if test_case.expected_output else None
                if test_case.raises:
                    with self.assertRaises(pydantic.ValidationError):
                        habapp_rules.actors.config.shading.SlatValueParameter(elevation_slat_characteristic=input_conf, elevation_slat_characteristic_summer=input_conf)
                else:
                    config = habapp_rules.actors.config.shading.SlatValueParameter(elevation_slat_characteristic=input_conf, elevation_slat_characteristic_summer=input_conf)

                    self.assertEqual(output, config.elevation_slat_characteristic)
                    self.assertEqual(output, config.elevation_slat_characteristic_summer)



"""Test notification rules."""

import unittest.mock

import HABApp.openhab.items
import multi_notifier.connectors.connector_mail
import multi_notifier.connectors.connector_telegram

import habapp_rules.system.notification
import tests.helper.oh_item
import tests.helper.test_case_base
from habapp_rules.system.config.notification import NotificationConfig, NotificationItems, NotificationParameter


class TestNotification(tests.helper.test_case_base.TestCaseBase):
    """Test class for notification."""

    def setUp(self) -> None:
        """Set up test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_String", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch", None)

        self._mail_mock = unittest.mock.MagicMock(spec=multi_notifier.connectors.connector_mail.Mail)
        self._telegram_mock = unittest.mock.MagicMock(spec=multi_notifier.connectors.connector_telegram.Telegram)

        self._mail_rule = habapp_rules.system.notification.SendStateChanged(
            NotificationConfig(items=NotificationItems(target_item=HABApp.openhab.items.OpenhabItem.get_item("Unittest_String")), parameter=NotificationParameter(notify_connector=self._mail_mock, recipients="mock@mail.de"))
        )
        self._telegram_rule = habapp_rules.system.notification.SendStateChanged(
            NotificationConfig(items=NotificationItems(target_item=HABApp.openhab.items.OpenhabItem.get_item("Unittest_Switch")), parameter=NotificationParameter(notify_connector=self._telegram_mock, recipients="mock_id"))
        )

    def test_state_changed(self) -> None:
        """Test state changed."""
        self._mail_mock.send_message.assert_not_called()
        self._telegram_mock.send_message.assert_not_called()

        tests.helper.oh_item.item_state_change_event("Unittest_String", "New value")
        self._mail_mock.send_message.assert_called_once_with("mock@mail.de", "Unittest_String changed from None to New value", subject="Unittest_String changed")

        tests.helper.oh_item.item_state_change_event("Unittest_String", "Even never value")
        self._mail_mock.send_message.assert_called_with("mock@mail.de", "Unittest_String changed from New value to Even never value", subject="Unittest_String changed")

        tests.helper.oh_item.item_state_change_event("Unittest_Switch", "ON")
        self._telegram_mock.send_message.assert_called_once_with("mock_id", "Unittest_Switch changed from None to ON")

        tests.helper.oh_item.item_state_change_event("Unittest_Switch", "OFF")
        self._telegram_mock.send_message.assert_called_with("mock_id", "Unittest_Switch changed from ON to OFF")



"""Rules to handle sun sensors."""

import logging

import HABApp

import habapp_rules.common.config.filter
import habapp_rules.common.filter
import habapp_rules.common.hysteresis
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.sensors.config.sun
from habapp_rules.core.helper import send_if_different
from habapp_rules.system import PresenceState

LOGGER = logging.getLogger(__name__)


class _SensorBase(HABApp.Rule):
    """Base class for sun sensors."""

    def __init__(self, config: habapp_rules.sensors.config.sun.BrightnessConfig | habapp_rules.sensors.config.sun.TemperatureDifferenceConfig, item_input: HABApp.openhab.items.NumberItem) -> None:
        """Init of base class for sun sensors.

        Args:
            config: config for sun sensor
            item_input: item for input value (brightness or temperature difference)
        """
        self._config = config

        # init HABApp Rule
        HABApp.Rule.__init__(self)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.output.name)

        # init exponential filter
        name_input_exponential_filtered = f"H_{item_input.name.removeprefix('H_')}_filtered"
        habapp_rules.core.helper.create_additional_item(name_input_exponential_filtered, "Number", name_input_exponential_filtered.replace("_", " "), config.parameter.filtered_signal_groups)
        item_input_filtered = HABApp.openhab.items.NumberItem.get_item(name_input_exponential_filtered)

        exponential_filter_config = habapp_rules.common.config.filter.ExponentialFilterConfig(
            items=habapp_rules.common.config.filter.ExponentialFilterItems(raw=item_input, filtered=item_input_filtered),
            parameter=habapp_rules.common.config.filter.ExponentialFilterParameter(tau=config.parameter.filter_tau, instant_increase=config.parameter.filter_instant_increase, instant_decrease=config.parameter.filter_instant_decrease),
        )
        habapp_rules.common.filter.ExponentialFilter(exponential_filter_config)

        # attributes
        self._hysteresis_switch = habapp_rules.common.hysteresis.HysteresisSwitch(config.threshold, config.parameter.hysteresis, return_bool=False)

        # callbacks
        item_input_filtered.listen_event(self._cb_input_filtered, HABApp.openhab.events.ItemStateChangedEventFilter())
        if config.items.threshold is not None:
            config.items.threshold.listen_event(self._cb_threshold, HABApp.openhab.events.ItemStateChangedEventFilter())

    def _send_output(self, new_value: str) -> None:
        """Send output if different.

        Args:
            new_value: new value which should be sent
        """
        if new_value != self._config.items.output.value:
            self._config.items.output.oh_send_command(new_value)
            self._instance_logger.debug(f"Set output '{self._config.items.output.name}' to {new_value}")

    def _cb_input_filtered(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the filtered input value changed.

        Args:
            event: trigger event
        """
        value = self._hysteresis_switch.get_output(event.value)
        self._send_output(value)

    def _cb_threshold(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the threshold changed.

        Args:
            event: trigger event
        """
        self._hysteresis_switch.set_threshold_on(event.value)


class SensorBrightness(_SensorBase):
    """Rules class to set sun protection depending on brightness level.

    # Items:
    Number    brightness                    "Current brightness"               {channel="..."}
    Number    brightness_threshold          "Brightness threshold"
    Switch    sun_protection_brightness     "Sun protection brightness"        {channel="..."}

    # Config:
    config = habapp_rules.sensors.config.sun.BrightnessConfig(
            items=habapp_rules.sensors.config.sun.BrightnessItems(
                    brightness="brightness",
                    output="sun_protection_brightness",
                    brightness_threshold="brightness_threshold"
            )
    )

    # Rule init:
    habapp_rules.sensors.sun.SensorBrightness(config)
    """

    def __init__(self, config: habapp_rules.sensors.config.sun.BrightnessConfig) -> None:
        """Init of sun sensor which takes a brightness value.

        Args:
            config: config for the sun sensor which is using brightness
        """
        _SensorBase.__init__(self, config, config.items.brightness)


class SensorTemperatureDifference(_SensorBase):
    """Rules class to set sun protection depending on temperature difference. E.g. temperature in the sun / temperature in the shadow.

    # Items:
    Number    temperature_sun               "Temperature sun"               {channel="..."}
    Number    temperature_shadow            "Temperature shadow"            {channel="..."}
    Number    temperature_threshold         "Temperature threshold"
    Switch    sun_protection_temperature    "Sun protection temperature"    {channel="..."}

    # Config:
    config = habapp_rules.sensors.config.sun.TemperatureDifferenceConfig(
            items=habapp_rules.sensors.config.sun.TemperatureDifferenceItems(
                    temperatures=["temperature_sun", "temperature_shadow"],
                    output="sun_protection_temperature",
                    threshold="temperature_threshold"
            )
    )

    # Rule init:
    habapp_rules.sensors.sun.SensorTempDiff(config)
    """

    def __init__(self, config: habapp_rules.sensors.config.sun.TemperatureDifferenceConfig) -> None:
        """Init of sun sensor which takes a two or more temperature values (one in the sun and one in the shadow).

        Args:
            config: config for the sun sensor which is using temperature items
        """
        self._config = config
        name_temperature_diff = f"H_Temperature_diff_for_{config.items.output.name}"
        habapp_rules.core.helper.create_additional_item(name_temperature_diff, "Number", name_temperature_diff.replace("_", " "), config.parameter.filtered_signal_groups)
        self._item_temp_diff = HABApp.openhab.items.NumberItem.get_item(name_temperature_diff)

        _SensorBase.__init__(self, config, self._item_temp_diff)

        # callbacks
        for temperature_item in self._config.items.temperatures:
            temperature_item.listen_event(self._cb_temperature, HABApp.openhab.events.ItemStateChangedEventFilter())

        # calculate temperature difference
        self._cb_temperature(None)

    def _cb_temperature(self, _: HABApp.openhab.events.ItemStateChangedEvent | None) -> None:
        """Callback, which is triggered if a temperature value changed."""
        filtered_items = [itm for itm in habapp_rules.core.helper.filter_updated_items(self._config.items.temperatures, self._config.parameter.ignore_old_values_time) if itm.value is not None]
        if len(filtered_items) < 2:  # noqa: PLR2004
            return
        value_min = min(item.value for item in filtered_items)
        value_max = max(item.value for item in filtered_items)

        self._item_temp_diff.oh_send_command(value_max - value_min)


class SunPositionFilter(HABApp.Rule):
    """Rules class to filter a switch state depending on the sun position. This can be used to only close the blinds of a window, if the sun hits the window.

    # Items:
    Number    sun_azimuth           "Sun Azimuth"                       {channel="astro..."}
    Number    sun_elevation         "Sun Elevation"                     {channel="astro..."}
    Switch    sun_shining           "Sun is shining"
    Switch    sun_hits_window       "Sun hits window"

    # Config:
    config = habapp_rules.sensors.config.sun.SunPositionConfig(
            items=habapp_rules.sensors.config.sun.SunPositionItems(
                    azimuth="sun_azimuth",
                    elevation="sun_elevation",
                    input="sun_shining",
                    output="sun_hits_window"
            ),
            parameter=habapp_rules.sensors.config.sun.SunPositionParameter(
                    sun_position_window=habapp_rules.sensors.config.sun.SunPositionWindow(40, 120)
            )
    )

    # Rule init:

    habapp_rules.sensors.sun.SunPositionFilter(config)
    """

    def __init__(self, config: habapp_rules.sensors.config.sun.SunPositionConfig) -> None:
        """Init of sun position filter.

        Args:
            config: config for the sun position filter
        """
        self._config = config

        # init HABApp Rule
        HABApp.Rule.__init__(self)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.output.name)

        # callbacks
        config.items.azimuth.listen_event(self._update_output, HABApp.openhab.events.ItemStateChangedEventFilter())  # listen_event for elevation is not needed because elevation and azimuth is updated together
        config.items.input.listen_event(self._update_output, HABApp.openhab.events.ItemStateChangedEventFilter())

        self._update_output(None)

    def _sun_in_window(self, azimuth: float, elevation: float) -> bool:
        """Check if the sun is in the 'sun window' where it hits the target.

        Args:
            azimuth: azimuth of the sun
            elevation: elevation of the sun

        Returns:
            True if the sun hits the target, else False
        """
        sun_in_window = False

        if any(window.azimuth_min <= azimuth <= window.azimuth_max and window.elevation_min <= elevation <= window.elevation_max for window in self._config.parameter.sun_position_windows):
            sun_in_window = True

        return sun_in_window

    def _update_output(self, _: HABApp.openhab.events.ItemStateChangedEvent | None) -> None:
        """Callback, which is triggered if the sun position or input changed."""
        azimuth = self._config.items.azimuth.value
        elevation = self._config.items.elevation.value

        if azimuth is None or elevation is None:
            self._instance_logger.warning(f"Azimuth or elevation is None -> will set output to input. azimuth = {azimuth} | elevation = {elevation}")
            filter_output = self._config.items.input.value
        elif self._config.items.input.value in {"OFF", None}:
            filter_output = "OFF"
        else:
            filter_output = "ON" if self._sun_in_window(azimuth, elevation) else "OFF"

        if filter_output != self._config.items.output.value:
            self._config.items.output.oh_send_command(filter_output)


class WinterFilter(HABApp.Rule):
    """Rule to filter the sun sensor depending on heating and presence state.

    # Items:
    Switch      sun             "Sun is shining"
    Switch      heating_active  "Heating is active"
    Switch      sun_filtered    "Sun filtered"

    # Config:
    config = habapp_rules.sensors.config.sun.WinterFilterConfig(
        items=habapp_rules.sensors.config.sun.WinterFilterItems(
            sun="sun",
            heating_active="heating_active",
            output="sun_filtered",
        )
    )

    # Rule init:
    habapp_rules.sensors.sun.WinterFilter(config)
    """

    def __init__(self, config: habapp_rules.sensors.config.sun.WinterFilterConfig) -> None:
        """Init of sun position filter.

        Args:
            config: config for the sun position filter
        """
        HABApp.Rule.__init__(self)
        self._config = config

        # callbacks
        config.items.sun.listen_event(self._cb_sun, HABApp.openhab.events.ItemStateChangedEventFilter())
        if config.items.heating_active is not None:
            config.items.heating_active.listen_event(self._cb_heating, HABApp.openhab.events.ItemStateChangedEventFilter())
        if config.items.presence_state is not None:
            config.items.presence_state.listen_event(self.cb_presence_state, HABApp.openhab.events.ItemStateChangedEventFilter())

        self._check_conditions_and_set_output()

    def _check_conditions_and_set_output(self) -> None:
        """Check conditions and set output.

        The output will be on, if the sun is up, the heating is off and somebody is at home.
        """
        heating_on = self._config.items.heating_active.is_on()
        absence = self._config.items.presence_state.value != PresenceState.PRESENCE.value if self._config.items.presence_state is not None else True

        target_state = self._config.items.sun.is_on() and (not heating_on or not absence)
        send_if_different(self._config.items.output, "ON" if target_state else "OFF")

    def _cb_sun(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:  # noqa: ARG002
        """Callback which is triggered if sun state changed.

        Args:
            event: original trigger event
        """
        self._check_conditions_and_set_output()

    def _cb_heating(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:  # noqa: ARG002
        """Callback which is triggered if heating state changed.

        Args:
            event: original trigger event
        """
        self._check_conditions_and_set_output()

    def cb_presence_state(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:  # noqa: ARG002
        """Callback which is triggered if presence_state changed.

        Args:
            event: original trigger event
        """
        self._check_conditions_and_set_output()
